# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import io
import math
import os
import struct
import sys
import typing as t
from datetime import date, timedelta

from ..abstract import AbstractDirectoryEntry, AbstractFile
from ..commons import (
    BLOCK_SIZE,
    DECTAPE,
    READ_FILE_FULL,
    bytes_to_word,
    filename_match,
    pad_words,
)
from ..device.abstract import AbstractDevice
from ..uic import ANY_UIC, DEFAULT_UIC, UIC
from .abstract import AbstractRXBlockFilesystem
from .rad50 import asc_to_rad50_word, rad50_word_to_asc
from .rt11fs import rt11_canonical_filename

__all__ = [
    "DOS11DirectoryEntry",
    "DOS11File",
    "DOS11Filesystem",
    "date_to_dos11",
    "dos11_canonical_filename",
    "dos11_get_file_type_id",
    "dos11_split_fullname",
    "dos11_to_date",
]

MFD1_BLOCK = 1  # Master File Directory block #1
UFD_ENTRIES = 28  # Number of User File Directory entries in a block
LINKED_FILE_BLOCK_SIZE = BLOCK_SIZE - 2  # Linked file block size in bytes (510)
DECTAPE_MFD1_BLOCK = 0o100  # DECtape Master File Directory block #1
DECTAPE_MFD2_BLOCK = 0o101  # DECtape Master File Directory block #2
DECTAPE_UFD1_BLOCK = 0o102  # DECtape User File Directory block #1
DECTAPE_UFD2_BLOCK = 0o103  # DECtape User File Directory block #2
DECTAPE_BITMAP_BLOCK = 0o104  # DECtape bitmap block
DECTAPE_BLOCKS = 576  # Number of blocks on a DECtape
BITMAP_HEADER_SIZE = 4  # Size of the bitmap header in words
MAX_WORDS_PER_BITMAP = 60  # Maximum number of words in a bitmap block
WORDS_PER_BLOCK = 256  # Number of words per block
MFD_ENTRY_SIZE = 4  # MFD entry size in words
UFD_ENTRY_SIZE = 9  # UFD entry size in words
DEFAULT_PROTECTION_CODE = 0o233
DEFAULT_INTERLAVE_FACTOR = 1

# File types
LINKED_FILE_TYPE = 0
CONTIGUOUS_FILE_TYPE = 32768

FILE_TYPES = {
    LINKED_FILE_TYPE: "NOCONTIGUOUS",
    CONTIGUOUS_FILE_TYPE: "CONTIGUOUS",
}


def dos11_get_file_type_id(file_type: t.Optional[str], default: int = LINKED_FILE_TYPE) -> int:
    """
    Get the file type id from a string
    """
    if not file_type:
        return default
    file_type = file_type.upper()
    for file_id, file_str in FILE_TYPES.items():
        if file_str == file_type:
            return file_id
    raise Exception("?KMON-F-Invalid file type specified with option")


def dos11_to_date(val: int) -> t.Optional[date]:
    """
    Translate DOS-11 date to Python date
    """
    if val == 0:
        return None
    val = val & 0o77777  # low 15 bits only
    year = val // 1000 + 1970  # encoded year
    doy = val % 1000  # encoded day of year
    try:
        return date(year, 1, 1) + timedelta(days=doy - 1)
    except:
        return None


def date_to_dos11(val: date) -> int:
    """
    Translate Python date to DOS-11 date
    """
    if val is None:
        return 0
    # Calculate the number of years since 1970
    year = val.year - 1970
    # Calculate the day of the year
    doy = (val - date(val.year, 1, 1)).days + 1
    # Combine into DOS-11 format
    return ((year * 1000) + doy) & 0o77777


def dos11_canonical_filename(fullname: str, wildcard: bool = False) -> str:
    try:
        if "[" in fullname:
            uic: t.Optional[UIC] = UIC.from_str(fullname)
            fullname = fullname.split("]", 1)[1]
        else:
            uic = None
    except Exception:
        uic = None
    if fullname:
        fullname = rt11_canonical_filename(fullname, wildcard=wildcard)
    return f"{uic or ''}{fullname}"


def dos11_split_fullname(uic: UIC, fullname: t.Optional[str], wildcard: bool = True) -> t.Tuple[UIC, t.Optional[str]]:
    if fullname:
        if "[" in fullname:
            try:
                uic = UIC.from_str(fullname)
                fullname = fullname.split("]", 1)[1]
            except Exception:
                return uic, fullname
        if fullname:
            fullname = rt11_canonical_filename(fullname, wildcard=wildcard)
    return uic, fullname


class DOS11Bitmap:
    """
    DOS-11 Bitmap

        +-------------------------------------+
      0 |     Link to next bitmap block       |
        +-------------------------------------+
      2 |         Map block number            |    sequential number of the bitmap block
        +-------------------------------------+
      4 |     Number of words in the map      |    constant for all the bitmap blocks
        +-------------------------------------+
      8 |   Link to the first bitmap block    |
        +-------------------------------------+
     10 |                                     |
        /          Map for blocks             /
    127 |                                     |
        +-------------------------------------+
    128 |                                     |
        /             Not used                /
    255 |                                     |
        +-------------------------------------+

    Disk Operating System Monitor - System Programmers Manual, Pag 138, 203
    http://www.bitsavers.org/pdf/dec/pdp11/dos-batch/DEC-11-OSPMA-A-D_PDP-11_DOS_Monitor_V004A_System_Programmers_Manual_May72.pdf
    """

    fs: "DOS11Filesystem"
    blocks: t.List[int]  # Bitmap block numbers
    num_of_words: int  # Number of words in each bitmap block
    bitmaps: t.List[int]

    def __init__(self, fs: "DOS11Filesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "DOS11Filesystem", first_bitmap_block: int) -> "DOS11Bitmap":
        """
        Read the bitmap blocks
        """
        self = DOS11Bitmap(fs)
        self.blocks = []
        self.bitmaps = []
        block_number = first_bitmap_block
        while block_number:
            # Read the bitmaps from the disk
            self.blocks.append(block_number)
            words = self.fs.read_words_block(block_number)
            if not words:
                raise OSError(errno.EIO, f"Failed to read block {block_number}")
            (
                block_number,  #      1 word  Next bitmap block number
                _,  #                 1 word  Map block number
                self.num_of_words,  # 1 word  Number of words of map
                _,  #                 1 word  First bitmap block number
            ) = words[:BITMAP_HEADER_SIZE]
            self.bitmaps.extend(words[BITMAP_HEADER_SIZE : BITMAP_HEADER_SIZE + self.num_of_words])
        return self

    @classmethod
    def new(cls, fs: "DOS11Filesystem", first_bitmap_block: int) -> "DOS11Bitmap":
        """
        Create a new bitmap
        """
        self = DOS11Bitmap(fs)
        number_of_blocks = self.fs.get_size() // BLOCK_SIZE
        total_num_of_words = math.ceil(number_of_blocks / 16)
        if total_num_of_words <= MAX_WORDS_PER_BITMAP:
            self.num_of_words = total_num_of_words  # Number of words in each bitmap block
            bitmap_blocks = 1  # Only one bitmap block needed
        else:
            self.num_of_words = MAX_WORDS_PER_BITMAP  # Maximum number of words in a bitmap block
            bitmap_blocks = math.ceil(total_num_of_words / self.num_of_words)
        self.blocks = list(range(first_bitmap_block, first_bitmap_block + bitmap_blocks))
        self.bitmaps = [0] * self.num_of_words * bitmap_blocks
        # Mark the bitmap blocks as used
        self.set_used(0)  # Block 0 is always used
        self.set_used(self.fs.mfd_block1)  # MFD block 1
        self.set_used(self.fs.mfd_block2)  # MFD block 2
        for block in self.blocks:
            self.set_used(block)
        for block in range(number_of_blocks, self.total_bits):
            self.set_used(block)
        self.write()
        return self

    def write(self) -> None:
        """
        Write the bitmap blocks
        """
        for bitmap_num in range(0, len(self.blocks)):
            next_block = self.blocks[bitmap_num + 1] if bitmap_num < len(self.blocks) - 1 else 0
            words = [
                next_block,  #        1 word  Next bitmap block number
                bitmap_num + 1,  #    1 word  Map block number
                self.num_of_words,  # 1 word  Number of words of map
                self.blocks[0],  #    1 word  First bitmap block number
            ]
            words += self.bitmaps[bitmap_num * self.num_of_words : (bitmap_num + 1) * self.num_of_words]
            words = pad_words(words, WORDS_PER_BLOCK)  # Fill the rest with zeros
            self.fs.write_words_block(self.blocks[bitmap_num], words)

    @property
    def total_bits(self) -> int:
        """
        Return the bitmap length in bit
        """
        return len(self.bitmaps) * 16

    def is_free(self, bit_index: int) -> bool:
        """
        Check if a block is free
        """
        int_index = bit_index // 16
        bit_position = bit_index % 16
        bit_value = self.bitmaps[int_index]
        return (bit_value & (1 << bit_position)) == 0

    def set_used(self, bit_index: int) -> None:
        """
        Mark a block as used
        """
        int_index = bit_index // 16
        bit_position = bit_index % 16
        self.bitmaps[int_index] |= 1 << bit_position

    def set_free(self, bit_index: int) -> None:
        """
        Mark a block as free
        """
        int_index = bit_index // 16
        bit_position = bit_index % 16
        self.bitmaps[int_index] &= ~(1 << bit_position)

    def find_contiguous_blocks(self, size: int) -> int:
        """
        Find contiguous blocks, return the first block number
        """
        current_run = 0
        start_index = -1
        for i in range(self.total_bits - 1, -1, -1):
            if self.is_free(i):
                if current_run == 0:
                    start_index = i
                current_run += 1
                if current_run == size:
                    return start_index - size + 1
            else:
                current_run = 0
        raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))

    def allocate(self, size: int, contiguous: bool = False) -> t.List[int]:
        """
        Allocate contiguous or sparse blocks
        """
        blocks = []
        if contiguous and size != 1:
            start_block = self.find_contiguous_blocks(size)
            for block in range(start_block, start_block + size):
                self.set_used(block)
                blocks.append(block)
        else:
            for block in range(0, self.total_bits):
                if self.is_free(block):
                    self.set_used(block)
                    blocks.append(block)
                if len(blocks) == size:
                    break
            if len(blocks) < size:
                raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        return blocks

    def used(self) -> int:
        """
        Count the number of used blocks
        """
        used = 0
        for block in self.bitmaps:
            used += block.bit_count()
        return used

    def free(self) -> int:
        """
        Count the number of free blocks
        """
        return len(self.bitmaps) * 16 - self.used()

    def __str__(self) -> str:
        free = self.free()
        used = self.used()
        return f"Free blocks: {free:<6} Used blocks: {used:<6}"


class DOS11File(AbstractFile):
    entry: "DOS11DirectoryEntry"
    closed: bool
    length: int  # Length in blocks
    contiguous: bool

    def __init__(self, entry: "DOS11DirectoryEntry"):
        self.entry = entry
        self.closed = False
        self.contiguous = entry.contiguous
        self.length = entry.length

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        Contiguous file block size is 512
        Linked file block size is 510
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.length
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.length
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        if self.contiguous:
            # Contiguous file
            return self.entry.ufd_block.fs.read_block(
                self.entry.start_block + block_number,
                number_of_blocks,
            )
        else:
            # Linked file
            seq = 0
            data = bytearray()
            next_block_number = self.entry.start_block
            while next_block_number != 0 and number_of_blocks:
                t = self.entry.ufd_block.fs.read_block(next_block_number)
                next_block_number = bytes_to_word(t, 0)
                if seq >= block_number:
                    data.extend(t[2:])
                    number_of_blocks -= 1
                seq += 1
            return bytes(data)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.length
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        if self.contiguous:
            # Contiguous file
            buffer += bytearray(BLOCK_SIZE * number_of_blocks - len(buffer))
            self.entry.ufd_block.fs.write_block(
                buffer,
                self.entry.start_block + block_number,
                number_of_blocks,
            )
        else:
            # Linked file
            seq = 0
            next_block_number = self.entry.start_block
            block_size = self.get_block_size()
            while next_block_number != 0 and number_of_blocks:
                t = self.entry.ufd_block.fs.read_block(next_block_number)
                if seq >= block_number:
                    t = t[:2] + buffer[:block_size]
                    t += bytearray(BLOCK_SIZE - len(t))
                    buffer = buffer[block_size:]
                    number_of_blocks -= 1
                    self.entry.ufd_block.fs.write_block(t, next_block_number)
                next_block_number = bytes_to_word(t, 0)
                seq += 1

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.length

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.get_length() * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE if self.contiguous else LINKED_FILE_BLOCK_SIZE

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class DOS11DirectoryEntry(AbstractDirectoryEntry):
    """
    User File Directory Entry

        +-------------------------------------+
     0  |               File                  |
     2  |               name                  |
        +-------------------------------------+
     4  |            Extension                |
        +-------------------------------------+
     6  |Type| Reserved |    Creation Date    |
        +-------------------------------------+
     8  |     Spare     | Lock | Usage count  |
        +-------------------------------------+
    10  |           Start block #             |
        +-------------------------------------+
    12  |        Length (# of blocks)         |
        +-------------------------------------+
    14  |            End block #              |
        +-------------------------------------+
    16  |     Spare     |   Protection code   |
        +-------------------------------------+

    Disk Operating System Monitor - System Programmers Manual, Pag 136, 202
    http://www.bitsavers.org/pdf/dec/pdp11/dos-batch/DEC-11-OSPMA-A-D_PDP-11_DOS_Monitor_V004A_System_Programmers_Manual_May72.pdf
    """

    ufd_block: "UserFileDirectoryBlock"
    uic: UIC = DEFAULT_UIC
    filename: str = ""
    extension: str = ""
    raw_creation_date: int = 0
    start_block: int = 0  # Block number of the first logical block
    length: int = 0  # Length in blocks
    end_block: int = 0  # Block number of the last logical block
    contiguous: bool = False  # Linked/contiguous file
    protection_code: int = 0  # System Programmers Manual, Pag 140
    usage_count: int = 0  # System Programmers Manual, Pag 136
    spare1: int = 0
    spare2: int = 0

    def __init__(self, ufd_block: "UserFileDirectoryBlock"):
        self.ufd_block = ufd_block
        self.uic = ufd_block.uic

    @classmethod
    def read(cls, ufd_block: "UserFileDirectoryBlock", words: t.List[int], position: int) -> "DOS11DirectoryEntry":
        # DOS Course Handouts, Pag 14
        # http://www.bitsavers.org/pdf/dec/pdp11/dos-batch/DOS_CourseHandouts.pdf
        self = DOS11DirectoryEntry(ufd_block)
        self.filename = rad50_word_to_asc(words[position]) + rad50_word_to_asc(words[position + 1])
        self.extension = rad50_word_to_asc(words[position + 2])
        self.raw_creation_date = words[position + 3]  # Type, Creation date
        self.spare1 = words[position + 4] >> 8  # Spare
        self.usage_count = words[position + 4] & 0xFF  # Lock, usage count
        self.start_block = words[position + 5]  # Block number of the first logical block
        self.length = words[position + 6]  # Length in blocks
        self.end_block = words[position + 7]  # Block number of the last logical block
        self.protection_code = words[position + 8] & 0xFF  # Protection code
        self.spare2 = words[position + 8] >> 8  # Spare
        if self.raw_creation_date & CONTIGUOUS_FILE_TYPE:
            self.contiguous = True
            self.raw_creation_date &= ~CONTIGUOUS_FILE_TYPE
        else:
            self.contiguous = False
        return self

    def to_words(self) -> t.List[int]:
        """
        Dump the directory entry to words
        """
        # Adjust raw_creation_date for contiguous files
        if self.contiguous:
            raw_creation_date = self.raw_creation_date | CONTIGUOUS_FILE_TYPE
        else:
            raw_creation_date = self.raw_creation_date & ~CONTIGUOUS_FILE_TYPE
        # Pack values into words
        return [
            asc_to_rad50_word(self.filename[:3]),  # File Name (1 word)
            asc_to_rad50_word(self.filename[3:6]),  # File Name (1 word)
            asc_to_rad50_word(self.extension),  # File Type (1 word)
            raw_creation_date,  # Type, Creation date (1 word)
            (self.spare1 << 8) | self.usage_count,  # Lock, usage count (1 word)
            self.start_block,  # Block number of the first logical block (1 word)
            self.length,  # Length in blocks (1 word)
            self.end_block,  # Block number of the last logical block (1 word)
            (self.spare2 << 8) | self.protection_code,  # Protection code (1 word)
        ]

    @property
    def is_empty(self) -> bool:
        return self.filename == "" and self.extension == ""

    @property
    def fullname(self) -> str:
        return f"{self.uic or ''}{self.filename}.{self.extension}"

    @property
    def basename(self) -> str:
        return f"{self.filename}.{self.extension}"

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.length

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.length * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE if self.contiguous else LINKED_FILE_BLOCK_SIZE

    @property
    def creation_date(self) -> t.Optional[date]:
        return dos11_to_date(self.raw_creation_date)

    def delete(self) -> bool:
        contiguous = self.contiguous
        start_block = self.start_block
        length = self.length
        # Write an empty User File Directory Entry
        self.raw_creation_date = 0
        self.usage_count = 0
        self.spare1 = 0
        self.start_block = 0
        self.length = 0
        self.end_block = 0
        self.protection_code = 0
        self.spare2 = 0
        self.filename = ""
        self.extension = ""
        self.contiguous = False
        self.ufd_block.write()
        # Free space
        bitmap = self.ufd_block.fs.read_bitmap()
        if contiguous:
            # Contiguous file
            for block_number in range(start_block, start_block + length):
                bitmap.set_free(block_number)
        else:
            # Linked file
            next_block_number = start_block
            while next_block_number != 0:
                bitmap.set_free(next_block_number)
                t = self.ufd_block.fs.read_block(next_block_number)
                next_block_number = bytes_to_word(t, 0)
        bitmap.write()
        return True

    def write(self) -> bool:
        """
        Write the directory entry
        """
        self.ufd_block.write()
        return True

    def open(self, file_mode: t.Optional[str] = None) -> DOS11File:
        """
        Open a file
        """
        return DOS11File(self)

    def __str__(self) -> str:
        return (
            f"{self.filename:<6}."
            f"{self.extension:<3}  "
            f"{self.uic.to_wide_str() if self.uic else '':<9}  "
            f"{self.creation_date or '          '} "
            f"{self.length:>6}{'C' if self.contiguous else ' '} "
            f"{self.start_block:6d} "
            f"{self.end_block:6d} "
            f"{self.protection_code:>6o} "
            f"{self.usage_count:>6o}"
        )

    def __repr__(self) -> str:
        return str(self)


class UserFileDirectoryBlock(object):
    """
    User File Directory Block

          +-------------------------------------+
       0  |          Link to next MFD           |
          +-------------------------------------+
       2  | UDF Entries                       1 |
          | .                                   |
          | .                                28 |
          +-------------------------------------+
     506  | Spare                               |
     512  | .                                   |
          +-------------------------------------+

    Each UFD Directory Block contains 28 entries, each entry is 9 words long.

    Disk Operating System Monitor - System Programmers Manual, Pag 136
    http://www.bitsavers.org/pdf/dec/pdp11/dos-batch/DEC-11-OSPMA-A-D_PDP-11_DOS_Monitor_V004A_System_Programmers_Manual_May72.pdf
    """

    # Block number of this user file directory block
    block_number = 0
    # Block number of the next user file directory block
    next_block_number = 0
    # User Identification Code
    uic: UIC = DEFAULT_UIC
    # User File Directory Block entries
    entries_list: t.List["DOS11DirectoryEntry"] = []

    def __init__(self, fs: "DOS11Filesystem", uic: UIC = DEFAULT_UIC):
        self.fs = fs
        self.uic = uic

    @classmethod
    def new(cls, fs: "DOS11Filesystem", uic: UIC, block_number: int) -> "UserFileDirectoryBlock":
        """
        Create a new empty User File Directory Block
        """
        self = UserFileDirectoryBlock(fs, uic)
        self.block_number = block_number
        self.next_block_number = 0
        self.entries_list = []
        for _ in range(1, UFD_ENTRIES * UFD_ENTRY_SIZE, UFD_ENTRY_SIZE):
            dir_entry = DOS11DirectoryEntry(self)
            self.entries_list.append(dir_entry)
        return self

    @classmethod
    def read(cls, fs: "DOS11Filesystem", uic: UIC, block_number: int) -> "UserFileDirectoryBlock":
        """
        Read a User File Directory Block from disk
        """
        self = UserFileDirectoryBlock(fs, uic)
        self.block_number = block_number
        words = self.fs.read_words_block(self.block_number)
        self.next_block_number = words[0]
        self.entries_list = []
        for position in range(1, UFD_ENTRIES * UFD_ENTRY_SIZE, UFD_ENTRY_SIZE):
            dir_entry = DOS11DirectoryEntry.read(self, words, position)
            self.entries_list.append(dir_entry)
        return self

    def write(self) -> None:
        """
        Write a User File Directory Block to disk
        """
        words = [self.next_block_number]  # Link to next block
        # Write each directory entry to the buffer
        for entry in self.entries_list:
            words += entry.to_words()
        words = pad_words(words, WORDS_PER_BLOCK)  # Fill the rest with zeros
        # Write the blocks to the disk
        self.fs.write_words_block(self.block_number, words)

    def get_empty_entry(self) -> t.Optional["DOS11DirectoryEntry"]:
        """
        Get the first empty directory entry
        """
        for entry in self.entries_list:
            if entry.is_empty:
                return entry
        return None

    def examine(self, options: t.Dict[str, t.Union[bool, str]]) -> str:
        include_all = bool(options.get("full", False))
        buf = io.StringIO()
        buf.write("\n*User File Directory Block\n")
        buf.write(f"UIC:                   {self.uic or ''}\n")
        buf.write(f"Block number:          {self.block_number}\n")
        buf.write(f"Next dir block:        {self.next_block_number}\n")
        header = False
        for i, x in enumerate(self.entries_list):
            if include_all or not x.is_empty:
                if not header:
                    buf.write("\nNum  File        UIC        Date       Length   Block    End   Code  Usage")
                    buf.write("\n---  ----        ---        ----       ------   -----    ---   ----  -----\n")
                    header = True
                buf.write(f"{i:02d}#  {x}\n")
        return buf.getvalue()

    def __str__(self) -> str:
        """
        String representation of the User File Directory Block
        """
        return self.examine({"full": True})


class MasterFileDirectoryEntry(AbstractDirectoryEntry):
    """
    Master File Directory Entry in the MFD block

          +-------------------------------------+
       0  |     Group code  |     User code     |
          +-------------------------------------+
       2  |          UFD start block #          |
          +-------------------------------------+
       4  |         # of words in UFD entry     |
          +-------------------------------------+
       6  |                 0                   |
          +-------------------------------------+

    Disk Operating System Monitor - System Programmers Manual, Pag 201
    https://bitsavers.org/pdf/dec/pdp11/dos-batch/DEC-11-OSPMA-A-D_PDP-11_DOS_Monitor_V004A_System_Programmers_Manual_May72.pdf
    """

    mfd_block: "AbstractMasterFileDirectoryBlock"
    uic: UIC = DEFAULT_UIC  # User Identification Code
    ufd_block: int = 0  # UFD start block
    num_words: int = 0  # num of words in UFD entry, always 9
    zero: int = 0  # always 0

    def __init__(self, mfd_block: "AbstractMasterFileDirectoryBlock"):
        self.mfd_block = mfd_block

    @classmethod
    def read(
        cls, mfd_block: "AbstractMasterFileDirectoryBlock", words: t.List[int], position: int
    ) -> "MasterFileDirectoryEntry":
        self = cls(mfd_block)
        self.uic = UIC.from_word(words[position])  # UIC
        self.ufd_block = words[position + 1]  # UFD start block
        self.num_words = words[position + 2]  # number of words in UFD entry
        self.zero = words[position + 3]  # always 0
        return self

    def to_words(self) -> t.List[int]:
        """
        Dump the directory entry to words
        """
        return [
            self.uic.to_word(),  # UIC
            self.ufd_block,  # UFD start block
            self.num_words,  # number of words in UFD entry
            self.zero,  # always 0
        ]

    def read_ufd_blocks(self) -> t.Iterator["UserFileDirectoryBlock"]:
        """Read User File Directory blocks"""
        next_block_number = self.ufd_block
        while next_block_number != 0:
            ufd_block = UserFileDirectoryBlock.read(self.fs, self.uic, next_block_number)
            next_block_number = ufd_block.next_block_number
            yield ufd_block

    def iterdir(
        self,
        pattern: t.Optional[str] = None,
        include_all: bool = False,
        wildcard: bool = False,
    ) -> t.Iterator["DOS11DirectoryEntry"]:
        for ufd_block in self.read_ufd_blocks():
            for entry in ufd_block.entries_list:
                if filename_match(entry.basename, pattern, wildcard):
                    if include_all or not entry.is_empty:
                        yield entry

    @property
    def is_empty(self) -> bool:
        return self.num_words == 0

    @property
    def fullname(self) -> str:
        return f"{self.uic}"

    @property
    def basename(self) -> str:
        return f"{self.uic}"

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return len(list(self.read_ufd_blocks()))

    def get_size(self) -> int:
        """
        Get entry size in bytes
        """
        return self.get_length() * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    def open(self, file_mode: t.Optional[str] = None) -> DOS11File:
        raise OSError(errno.EINVAL, "Invalid operation on directory")

    def delete(self) -> bool:
        # Delete all entries in the UFD
        for entry in self.iterdir():
            if not entry.delete():
                raise OSError(errno.EIO, os.strerror(errno.EIO))
        # Free space
        bitmap = self.mfd_block.fs.read_bitmap()
        bitmap.set_free(self.ufd_block)
        # Write an empty Master File Directory Entry
        self.uic = UIC(0, 0)
        self.ufd_block = 0
        self.num_words = 0
        self.zero = 0
        self.mfd_block.write()  # type: ignore
        # Write the bitmap
        bitmap.write()
        return True

    def write(self) -> bool:
        """
        Write the directory entry
        """
        raise OSError(errno.EINVAL, "Invalid operation on directory")

    @property
    def fs(self) -> "DOS11Filesystem":
        return self.mfd_block.fs

    def __str__(self) -> str:
        return f"{self.uic} ufd_block={self.ufd_block} num_words={self.num_words} zero={self.zero}"


class AbstractMasterFileDirectoryBlock:
    """
    DOS-11/XXDP+ Master File Directory Block
    """

    fs: "DOS11Filesystem"
    # Master File Directory Block entries
    entries_list: t.List["MasterFileDirectoryEntry"] = []


class MasterFileDirectoryBlock(AbstractMasterFileDirectoryBlock):
    """
    Master File Directory Block 2 - N

    MFD Block 2 - N:
          +-------------------------------------+
       0  |          Link to next MFD           |
          +-------------------------------------+
       2  | MFD Entries                       1 |
          | .                                   |
          | .                                28 |
          +-------------------------------------+

    Disk Operating System Monitor - System Programmers Manual, Pag 135
    http://www.bitsavers.org/pdf/dec/pdp11/dos-batch/DEC-11-OSPMA-A-D_PDP-11_DOS_Monitor_V004A_System_Programmers_Manual_May72.pdf
    """

    # Block number of this Master File Directory block
    block_number = 0
    # Block number of the next Master File Directory block
    next_block_number = 0

    def __init__(self, fs: "DOS11Filesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "DOS11Filesystem", block_number: int) -> "MasterFileDirectoryBlock":
        """
        Read a Master File Directory Block from disk
        """
        self = MasterFileDirectoryBlock(fs)
        self.block_number = block_number
        words = self.fs.read_words_block(self.block_number)
        self.next_block_number = words[0]  # link to next MFD
        self.entries_list = []
        for position in range(1, WORDS_PER_BLOCK - MFD_ENTRY_SIZE, MFD_ENTRY_SIZE):
            entry = MasterFileDirectoryEntry.read(self, words, position)
            self.entries_list.append(entry)
        return self

    @classmethod
    def new(cls, fs: "DOS11Filesystem", block_number: int) -> "MasterFileDirectoryBlock":
        """
        Create a new empty Master File Directory Block
        """
        self = MasterFileDirectoryBlock(fs)
        self.block_number = block_number
        self.next_block_number = 0
        self.entries_list = []
        for _ in range(1, WORDS_PER_BLOCK - MFD_ENTRY_SIZE, MFD_ENTRY_SIZE):
            entry = MasterFileDirectoryEntry(self)
            self.entries_list.append(entry)
        return self

    def write(self) -> None:
        """
        Write a Master File Directory Block to disk
        """
        words = [self.next_block_number]  # Link to next MFD block
        # Write each directory entry to the buffer
        for entry in self.entries_list:
            words += entry.to_words()
        words = pad_words(words, WORDS_PER_BLOCK)  # Fill the rest with zeros
        # Write the buffer to the disk
        self.fs.write_words_block(self.block_number, words)

    def get_empty_entry(self) -> t.Optional["MasterFileDirectoryEntry"]:
        """
        Get the first empty directory entry
        """
        for entry in self.entries_list:
            if entry.is_empty:
                return entry
        return None


class XXDPMasterFileDirectoryBlock(AbstractMasterFileDirectoryBlock):
    """
    XXDP Master File Directory

    XXDP has only one UFD in the MFD
    """

    def __init__(self, fs: "DOS11Filesystem"):
        self.fs = fs
        entry = MasterFileDirectoryEntry(self)
        entry.ufd_block = self.fs.xxdp_ufd_block
        entry.uic = self.fs.uic
        entry.num_words = UFD_ENTRY_SIZE
        self.entries_list = [entry]


class DOS11Filesystem(AbstractRXBlockFilesystem):
    """
    DOS-11/XXDP+ Filesystem

    General disk layout:

    Block
          +-------------------------------------+
    0     |            Bootstrap block          |
          +-------------------------------------+
    1     |             MFD Block #1            |
          +-------------------------------------+
    2     |             UFD Block #1            |
          +-------------------------------------+
          |           User linked files         |
          |           other UFD blocks          |
          |        User contiguous files        |
          +-------------------------------------+
    l-n   |             MFD Block #2            |
          +-------------------------------------+
    l-n-1 | Bitmap Block                      1 |
          | .                                   |
    l     | .                                 n |
          +--------------------------------------


    DOS-11 format - Pag 204
    http://www.bitsavers.org/pdf/dec/pdp11/dos-batch/DEC-11-OSPMA-A-D_PDP-11_DOS_Monitor_V004A_System_Programmers_Manual_May72.pdf

    DECtape format - Pag 206
    http://www.bitsavers.org/pdf/dec/pdp11/dos-batch/DEC-11-OSPMA-A-D_PDP-11_DOS_Monitor_V004A_System_Programmers_Manual_May72.pdf

    XXDP File Structure Guide - Pag 8
    https://raw.githubusercontent.com/rust11/xxdp/main/XXDP%2B%20File%20Structure.pdf
    """

    fs_name = "dos11"
    fs_description = "PDP-11 DOS-11/XXDP+"
    fs_platforms = ["pdp11"]

    uic: UIC  # current User Identification Code
    xxdp: bool = False  # MFD Variety #2 (XXDP+)
    dectape: bool = False  # DECtape format
    mfd_block1: int = MFD1_BLOCK  # Master File Directory Block #1
    mfd_block2: int = 0  # Master File Directory Block #2
    bitmap_block: int = 0  # Bitmap first block number
    interleave_factor: int = DEFAULT_INTERLAVE_FACTOR  # Interleave factor
    xxdp_ufd_block: int = 0  # XXDP+ UFD first block number

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        **kwargs: t.Union[bool, str],
    ) -> "DOS11Filesystem":
        self = cls(file_or_dev)
        self.uic = DEFAULT_UIC
        self.read_mfd()  # Read the Master File Directory
        if strict:
            # Check if the used blocks are in the bitmap
            blocks = [mfd.ufd_block for mfd in self.read_mfd_entries()]
            if not self.bitmap_block:
                raise OSError(errno.EIO, "Failed to read MFD block")
            bitmap = self.read_bitmap()
            for block in blocks:
                if bitmap.is_free(block):
                    raise OSError(errno.EIO, f"Block {block} is not in the bitmap")
            if bitmap.is_free(self.bitmap_block):
                raise OSError(errno.EIO, f"Block {self.bitmap_block} is not in the bitmap")
        return self

    @classmethod
    def initialize(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "DOS11Filesystem":
        """
        Create an empty filesystem
        """
        self = cls(file_or_dev)
        self.dectape = kwargs.get("device_type", "") == DECTAPE
        self.xxdp = False
        self.uic = DEFAULT_UIC
        self.interleave_factor = DEFAULT_INTERLAVE_FACTOR
        if self.dectape:
            self.mfd_block1 = DECTAPE_MFD1_BLOCK
            self.mfd_block2 = DECTAPE_MFD2_BLOCK
            self.bitmap_block = DECTAPE_BITMAP_BLOCK
        else:
            self.mfd_block1 = MFD1_BLOCK
            self.mfd_block2 = MFD1_BLOCK + 2
            self.bitmap_block = MFD1_BLOCK + 3

        # Create the bitmap
        bitmap = DOS11Bitmap.new(self, self.bitmap_block)

        # Create the Master File Directory Block #1
        words = [
            self.mfd_block2,  #        Master File Directory Block #2
            self.interleave_factor,  # Interleave factor
            self.bitmap_block,  #      Bitmap start block
        ] + bitmap.blocks  #           Bitmap blocks
        words = pad_words(words, WORDS_PER_BLOCK)  # Fill the rest with zeros
        self.write_words_block(self.mfd_block1, words)

        # Create the Master File Directory Block #2
        mfd_block = MasterFileDirectoryBlock.new(self, self.mfd_block2)
        mfd_block.write()

        # Create an User File Directory
        self.create_directory("[1,1]", {})
        return self

    def read_words_block(
        self,
        block_number: int,
    ) -> t.List[int]:
        """
        Read a 512 bytes block as 256 16bit words
        """
        data = self.read_block(block_number)
        if not data:
            raise OSError(errno.EIO, f"Failed to read block {block_number}")
        return list(struct.unpack_from("<256H", data))

    def write_words_block(
        self,
        block_number: int,
        words: t.List[int],
    ) -> None:
        """
        Write 256 16bit words as a 512 bytes block
        """
        data = struct.pack("<256H", *words)
        self.write_block(data, block_number)

    def read_mfd_entries(
        self,
        uic: UIC = ANY_UIC,
    ) -> t.Iterator["MasterFileDirectoryEntry"]:
        """
        Read Master File Directory entries
        """
        for mfd in self.read_mfd_blocks():
            for entry in mfd.entries_list:
                if not entry.is_empty and uic.match(entry.uic):  # Filter by UIC
                    yield entry

    def read_mfd_blocks(self) -> t.Iterator["AbstractMasterFileDirectoryBlock"]:
        """
        Read Master File Directory blocks
        """
        if self.mfd_block2 != 0:  # MFD Variety #1 (DOS-11)
            mfd_block = self.mfd_block2
            while mfd_block:
                mfd = MasterFileDirectoryBlock.read(self, mfd_block)
                mfd_block = mfd.next_block_number
                yield mfd
        else:  # MFD Variety #2 (XXDP+)
            yield XXDPMasterFileDirectoryBlock(self)

    def read_mfd(self) -> None:
        """
        Read Master File Directory Block 1

        MFD Block 1:

              +-------------------------------------+
              |        Block # of MFD Block 2       |
              +-------------------------------------+
              |           Interleave factor         |
              +-------------------------------------+
              |            Bitmap block #1          |
              +-------------------------------------+
              |            Bitmap block #2          |
              /                                     /
              |            Bitmap block #n          |
              +-------------------------------------+
              |                    0                |
              +-------------------------------------+
              |                                     |

        """
        # Check DECtape format
        try:
            words = self.read_words_block(DECTAPE_MFD1_BLOCK)
            self.mfd_block2 = words[0]  # Next MFD block
            self.interleave_factor = words[1]  # Interleave factor
            self.bitmap_block = words[2]  # Bitmap start block
            if self.mfd_block2 != DECTAPE_MFD2_BLOCK:
                raise OSError("Not a DECtape filesystem")
            words = self.read_words_block(self.mfd_block2)
            if words[0] != 0:  # 0, DECtape has only 2 MFD
                raise OSError("Not a DECtape filesystem")
            if words[2] != DECTAPE_UFD1_BLOCK:
                raise OSError("Not a DECtape filesystem")
            self.mfd_block1 = DECTAPE_MFD1_BLOCK
            self.dectape = True
        except OSError:
            words = self.read_words_block(MFD1_BLOCK)
            self.mfd_block1 = MFD1_BLOCK
            self.mfd_block2 = words[0]  # Next MFD block
            self.interleave_factor = words[1]  # Interleave factor
            self.bitmap_block = words[2]  # Bitmap start block
            self.dectape = False

        if self.mfd_block2 == 0:  # MFD Variety #2 (XXDP+)
            # Pag 10
            # https://raw.githubusercontent.com/rust11/xxdp/main/XXDP%2B%20File%20Structure.pdf
            self.xxdp = True
            # words[0] # Always 0
            self.xxdp_ufd_block = words[1]  # First UFD block
            # words[2] # Number of UFD blocks, not used in XXDP+
            self.bitmap_block = words[3]  # Bitmap start block

    def read_bitmap(self) -> DOS11Bitmap:
        bitmap = DOS11Bitmap.read(self, self.bitmap_block)
        return bitmap

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,  # expand directories
        wildcard: bool = True,
        uic: t.Optional[UIC] = None,
    ) -> t.Iterator["DOS11DirectoryEntry"]:
        if uic is None:
            uic = self.uic
        uic, filename_pattern = dos11_split_fullname(fullname=pattern, wildcard=wildcard, uic=uic)
        if pattern and not filename_pattern and not expand:
            # If expand is False, check if the pattern is an UIC
            try:
                uic = UIC.from_str(pattern)
                for mfd_block in self.read_mfd_blocks():
                    for entry in mfd_block.entries_list:
                        if not entry.is_empty and uic.match(entry.uic):
                            yield entry  # type: ignore
                return
            except Exception:
                pass
        for mfd in self.read_mfd_entries(uic=uic):
            yield from mfd.iterdir(pattern=filename_pattern, include_all=include_all, wildcard=wildcard)

    @property
    def entries_list(self) -> t.Iterator["DOS11DirectoryEntry"]:
        for mfd in self.read_mfd_entries(uic=self.uic):
            yield from mfd.iterdir()

    def get_file_entry(self, fullname: str) -> DOS11DirectoryEntry:
        """
        Get the directory entry for a file
        """
        fullname = dos11_canonical_filename(fullname)
        if not fullname:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)
        uic, basename = dos11_split_fullname(fullname=fullname, wildcard=False, uic=self.uic)
        try:
            return next(self.filter_entries_list(basename, wildcard=False, uic=uic))
        except StopIteration:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
        protection_code: int = DEFAULT_PROTECTION_CODE,
    ) -> None:
        """
        Write content to a file
        """
        block_size = BLOCK_SIZE if dos11_get_file_type_id(file_type) == CONTIGUOUS_FILE_TYPE else LINKED_FILE_BLOCK_SIZE
        number_of_blocks = int(math.ceil(len(content) * 1.0 / block_size))
        entry = self.create_file(
            fullname=fullname,
            number_of_blocks=number_of_blocks,
            creation_date=creation_date,
            file_type=file_type,
            protection_code=protection_code,
        )
        if entry is not None:
            f = DOS11File(entry)
            try:
                f.write_block(content, 0, entry.length)
            finally:
                f.close()

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
        protection_code: int = DEFAULT_PROTECTION_CODE,
    ) -> t.Optional[DOS11DirectoryEntry]:
        """
        Create a new file with a given length in number of blocks
        """
        contiguous = dos11_get_file_type_id(file_type) == CONTIGUOUS_FILE_TYPE
        # Delete the existing file
        try:
            self.get_file_entry(fullname).delete()
        except FileNotFoundError:
            pass
        # Get the MFD entry for the target UIC
        uic, basename = dos11_split_fullname(fullname=fullname, wildcard=False, uic=self.uic)
        try:
            mfd = next(self.read_mfd_entries(uic=uic))
        except Exception:
            raise NotADirectoryError
        # Allocate the space for the file
        bitmap = self.read_bitmap()
        blocks = bitmap.allocate(number_of_blocks, contiguous)
        # Create the directory entry
        new_entry = None
        for ufd_block in mfd.read_ufd_blocks():
            new_entry = ufd_block.get_empty_entry()
            if new_entry is not None:
                break
        if new_entry is None:
            # Allocate a new UFD block
            new_block_number = bitmap.allocate(1)[0]
            # Write the link to new new block in the old block
            ufd_block.next_block_number = new_block_number
            ufd_block.write()
            # Create a new UFD block
            ufd_block = UserFileDirectoryBlock.new(self, uic, new_block_number)
            new_entry = ufd_block.get_empty_entry()
            if new_entry is None:
                raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        try:
            filename, extension = basename.split(".", 1)  # type: ignore
        except Exception:
            filename = basename
            extension = ""
        new_entry.filename = filename
        new_entry.extension = extension
        new_entry.raw_creation_date = date_to_dos11(creation_date or date.today())
        new_entry.start_block = blocks[0]
        new_entry.length = number_of_blocks
        new_entry.end_block = blocks[-1]
        new_entry.contiguous = contiguous
        new_entry.protection_code = protection_code
        new_entry.ufd_block.write()
        # Write bitmap
        bitmap.write()
        # Write linked file
        if not contiguous:
            for i, block in enumerate(blocks):
                buffer = bytearray(BLOCK_SIZE)
                next_block_number = blocks[i + 1] if i + 1 < len(blocks) else 0
                struct.pack_into("<H", buffer, 0, next_block_number)
                self.write_block(buffer, block)
        return new_entry

    def create_directory(
        self,
        fullname: str,
        options: t.Dict[str, t.Union[bool, str]],
    ) -> t.Optional["MasterFileDirectoryEntry"]:
        """
        Create a User File Directory
        """
        if self.xxdp:
            raise OSError(errno.EINVAL, "Invalid operation on XXDP+ filesystem")
        try:
            uic = UIC.from_str(fullname)
        except Exception:
            raise OSError(errno.EINVAL, "Invalid UIC")
        # Check if the UIC already exists
        if list(self.read_mfd_entries(uic=uic)):
            raise FileExistsError(errno.EEXIST, os.strerror(errno.EEXIST))
        found = False
        mfd: "MasterFileDirectoryBlock"
        for mfd in self.read_mfd_blocks():  # type: ignore
            entry: MasterFileDirectoryEntry = mfd.get_empty_entry()  # type: ignore
            if entry is not None:
                found = True
                break
        if not found:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        # Create a new UFD block
        bitmap = self.read_bitmap()
        if self.dectape:
            if bitmap.is_free(DECTAPE_UFD1_BLOCK):
                blocks = [DECTAPE_UFD1_BLOCK, DECTAPE_UFD2_BLOCK]
                bitmap.set_used(DECTAPE_UFD1_BLOCK)
                bitmap.set_used(DECTAPE_UFD2_BLOCK)
            else:
                blocks = bitmap.allocate(2, contiguous=True)
        else:
            blocks = bitmap.allocate(2)
        bitmap.write()
        # Initialize the blocks with zeros
        self.write_words_block(blocks[0], pad_words([blocks[1]], WORDS_PER_BLOCK))
        self.write_words_block(blocks[1], [0] * WORDS_PER_BLOCK)
        # Write the new entry
        entry.uic = uic
        entry.ufd_block = blocks[0]
        entry.num_words = UFD_ENTRY_SIZE
        mfd.write()
        return entry

    def isdir(self, fullname: str) -> bool:
        return False

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        if options.get("uic"):
            # Listing of all UIC
            sys.stdout.write(f"{volume_id}:\n\n")
            for mfd in self.read_mfd_entries(uic=ANY_UIC):
                sys.stdout.write(f"{mfd.uic.to_wide_str()}\n")
            return
        files = 0
        blocks = 0
        i = 0
        uic, pattern = dos11_split_fullname(fullname=pattern, wildcard=True, uic=self.uic)
        if not options.get("brief"):
            if self.xxdp:
                sys.stdout.write("ENTRY# FILNAM.EXT        DATE          LENGTH  START\n")
            else:
                dt = date.today().strftime('%y-%b-%d').upper()
                sys.stdout.write(f"DIRECTORY {volume_id}: {uic}\n\n{dt}\n\n")
        for x in self.filter_entries_list(pattern, uic=uic, include_all=True, wildcard=True):
            if x.is_empty:
                continue
            i = i + 1
            fullname = x.is_empty and x.filename or "%-6s.%-3s" % (x.filename, x.extension)
            if options.get("brief"):
                # Lists only file names and file types
                sys.stdout.write(f"{fullname}\n")
                continue
            creation_date = x.creation_date and x.creation_date.strftime("%d-%b-%y").upper() or ""
            attr = "C" if x.contiguous else ""
            if self.xxdp:
                sys.stdout.write(f"{i:6} {fullname:>10s} {creation_date:>14s} {x.length:>10d}    {x.start_block:06o}\n")
            else:
                uic_str = x.uic.to_wide_str() if uic.has_wildcard else ""
                sys.stdout.write(
                    f"{fullname:>10s} {x.length:>5d}{attr:1} {creation_date:>9s} <{x.protection_code:03o}> {uic_str}\n"
                )
            blocks += x.length
            files += 1
        if options.get("brief") or self.xxdp:
            return
        sys.stdout.write("\n")
        sys.stdout.write(f"TOTL BLKS: {blocks:5}\n")
        sys.stdout.write(f"TOTL FILES: {files:4}\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if options.get("bitmap"):
            # Display the bitmap
            bitmap = self.read_bitmap()
            for i in range(0, bitmap.total_bits):
                sys.stdout.write(f"{i:>4d} {'[ ]' if bitmap.is_free(i) else '[X]'}  ")
                if i % 16 == 15:
                    sys.stdout.write("\n")
            sys.stdout.write(f"\nUsed blocks: {bitmap.used()}\n")
        elif arg:
            # Display the file entry
            entries = self.filter_entries_list(arg, wildcard=True)
            for entry in entries:
                sys.stdout.write(f"UIC:                      {entry.uic.to_wide_str()}\n")
                sys.stdout.write(f"Filename:                 {entry.filename}\n")
                sys.stdout.write(f"Extension:                {entry.extension}\n")
                sys.stdout.write(f"Creation date:            {entry.creation_date}\n")
                sys.stdout.write(f"Length in blocks:         {entry.length}\n")
                sys.stdout.write(f"Start block:              {entry.start_block}\n")
                sys.stdout.write(f"End block:                {entry.end_block}\n")
                sys.stdout.write(f"Contiguous:               {'Y' if entry.contiguous else 'N'}\n")
                sys.stdout.write(f"Protection code:          {entry.protection_code:03o}\n")
                sys.stdout.write(f"Usage count:              {entry.usage_count}\n")
        else:
            bitmap = self.read_bitmap()
            sys.stdout.write(f"Free blocks:       {bitmap.free():>8}\n")
            sys.stdout.write(f"Used blocks:       {bitmap.used():>8}\n")
            sys.stdout.write(f"Is DECtape:               {'Y' if self.dectape else 'N'}\n")
            sys.stdout.write(f"Is XXDP+:                 {'Y' if self.xxdp else 'N'}\n")
            sys.stdout.write(f"MFD block #1:      {self.mfd_block1:>8}\n")
            sys.stdout.write(f"MFD block #2:      {self.mfd_block2:>8}\n")
            sys.stdout.write(f"Bitmap block:      {self.bitmap_block:>8}\n")
            sys.stdout.write(f"Bitmap blocks:     {len(bitmap.blocks):>8}\n")
            sys.stdout.write(f"Bitmap words:      {bitmap.num_of_words:>8}\n")
            sys.stdout.write(f"Interleave factor: {self.interleave_factor:>8}\n")
            for mfd in self.read_mfd_entries():
                for ufd_block in mfd.read_ufd_blocks():
                    sys.stdout.write(ufd_block.examine(options=options))

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        if self.dectape:
            return DECTAPE_BLOCKS * BLOCK_SIZE
        else:
            return self.dev.get_size()

    def chdir(self, fullname: str) -> bool:
        """
        Change the current User Identification Code
        """
        try:
            self.uic = UIC.from_str(fullname)
            return True
        except Exception:
            return False

    def get_pwd(self) -> str:
        """
        Get the current User Identification Code
        """
        return str(self.uic)

    def get_types(self) -> t.List[str]:
        """
        Get the list of the supported file types
        """
        return list(FILE_TYPES.values())



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import glob
import io
import math
import os
import stat
import sys
import typing as t
from datetime import date, datetime

from .abstract import AbstractDirectoryEntry, AbstractFile, AbstractFilesystem
from .commons import BLOCK_SIZE, READ_FILE_FULL
from .device.abstract import AbstractDevice

__all__ = [
    "NativeFile",
    "NativeDirectoryEntry",
    "NativeFilesystem",
]


def format_size(size: float) -> str:
    for unit in ("", "K", "M", "G", "T"):
        if abs(size) < 1000 or unit == "T":
            break
        size = size / 1000
    return f"{size:3.1f}{unit}" if unit else str(int(size))


class NativeFile(AbstractFile):

    f: t.Union[io.BufferedReader, io.BufferedRandom]

    def __init__(self, filename: str):
        self.filename = os.path.abspath(filename)
        try:
            self.f = open(filename, mode="rb+")
            self.readonly = False
        except OSError:
            self.f = open(filename, mode="rb")
            self.readonly = True
        self.size = os.path.getsize(filename)

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            self.f.seek(0)  # not thread safe...
            return self.f.read()
        elif block_number < 0 or number_of_blocks < 0:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        else:
            position = block_number * BLOCK_SIZE
            self.f.seek(position)  # not thread safe...
            return self.f.read(number_of_blocks * BLOCK_SIZE)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if block_number < 0 or number_of_blocks < 0:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        elif self.readonly:
            raise OSError(errno.EROFS, os.strerror(errno.EROFS))
        else:
            self.f.seek(block_number * BLOCK_SIZE)  # not thread safe...
            self.f.write(buffer[0 : number_of_blocks * BLOCK_SIZE])

    def truncate(self, size: t.Optional[int] = None) -> None:
        """
        Resize the file to the given number of bytes.
        If the size is not specified, the current position will be used.
        """
        self.f.truncate(size)
        if size is not None and self.current_position > size:
            self.current_position = size

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.size

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    def close(self) -> None:
        """
        Close the file
        """
        self.f.close()

    def __str__(self) -> str:
        return self.filename


class NativeDirectoryEntry(AbstractDirectoryEntry):

    def __init__(self, fullname: str):
        self.native_fullname = fullname
        self.filename = os.path.basename(fullname)
        self.filename, self.extension = os.path.splitext(self.filename)
        if self.extension.startswith("."):
            self.extension = self.filename[1:]
        self.stat = os.stat(fullname)
        self.length = self.stat.st_size  # length in bytes

    @property
    def creation_date(self) -> date:
        return datetime.fromtimestamp(self.stat.st_mtime)

    @property
    def fullname(self) -> str:
        return self.native_fullname

    @property
    def basename(self) -> str:
        return os.path.basename(self.native_fullname)

    @property
    def is_regular_file(self) -> bool:
        """
        Check if the entry is a regular file
        """
        return stat.S_ISREG(self.stat.st_mode)

    @property
    def is_directory(self) -> bool:
        """
        Check if the entry is a directory
        """
        return stat.S_ISDIR(self.stat.st_mode)

    @property
    def is_link(self) -> bool:
        """
        Check if the entry is a symbolic link
        """
        return stat.S_ISLNK(self.stat.st_mode)

    @property
    def entry_type(self) -> t.Optional[str]:
        """
        Entry type
        """
        mode = self.stat.st_mode
        if stat.S_ISREG(mode):
            return "FILE"
        elif stat.S_ISDIR(mode):
            return "DIRECTORY"
        elif stat.S_ISLNK(mode):
            return "LINK"
        elif stat.S_ISFIFO(mode):
            return "FIFO"
        elif stat.S_ISSOCK(mode):
            return "SOCKET"
        elif stat.S_ISCHR(mode):
            return "CHAR DEV"
        elif stat.S_ISBLK(mode):
            return "BLOCK DEV"
        else:
            return None

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return int(math.ceil(self.get_size() / self.get_block_size()))

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.stat.st_size

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    def delete(self) -> bool:
        """
        Delete the directory entry
        """
        try:
            if self.is_directory:
                os.rmdir(self.native_fullname)
            else:
                os.unlink(self.native_fullname)
            return True
        except:
            return False

    def write(self) -> bool:
        """
        Write the directory entry
        """
        raise OSError(errno.EINVAL, "Invalid operation on native filesystem")

    def open(self, file_mode: t.Optional[str] = None) -> NativeFile:
        """
        Open a file
        """
        return NativeFile(self.fullname)

    def __str__(self) -> str:
        return f"{self.fullname:<11} {self.creation_date or '':<6} length: {self.length:>6}"


class NativeFilesystem(AbstractFilesystem):

    @classmethod
    def mount(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "NativeFilesystem":
        if not isinstance(file_or_dev, NativeFile):
            raise OSError(errno.EIO, "Not a native file")
        return cls(base=file_or_dev.filename)

    def __init__(self, base: t.Optional[str] = None):
        self.base = base or "/"
        if not base:
            self.pwd = os.getcwd()
        elif os.getcwd().startswith(base):
            self.pwd = os.getcwd()[len(base) :]
        else:
            self.pwd = os.path.sep

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
    ) -> t.Iterator["NativeDirectoryEntry"]:
        if not pattern:
            for filename in os.listdir(os.path.join(self.base, self.pwd)):
                try:
                    v = NativeDirectoryEntry(os.path.join(self.base, self.pwd, filename))
                except:
                    v = None
                if v is not None:
                    yield v
        else:
            if not pattern.startswith("/") and not pattern.startswith("\\"):
                pattern = os.path.join(self.base, self.pwd, pattern)
            if os.path.isdir(pattern):
                if not expand:  # don't expand directories
                    yield NativeDirectoryEntry(pattern)
                    return
                pattern = os.path.join(pattern, "*")
            for filename in glob.glob(pattern):
                try:
                    v = NativeDirectoryEntry(filename)
                except:
                    v = None
                if v is not None:
                    yield v

    @property
    def entries_list(self) -> t.Iterator["NativeDirectoryEntry"]:
        dir = self.pwd
        for filename in os.listdir(dir):
            yield NativeDirectoryEntry(os.path.join(dir, filename))

    def get_file_entry(self, fullname: str) -> "NativeDirectoryEntry":
        """
        Get the file entry for a given path
        """
        if not fullname.startswith("/") and not fullname.startswith("\\"):
            fullname = os.path.join(self.pwd, fullname)
        return NativeDirectoryEntry(fullname)

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        """
        Write content to a file
        """
        if not fullname.startswith("/") and not fullname.startswith("\\"):
            fullname = os.path.join(self.pwd, fullname)
        with open(fullname, "wb") as f:
            f.write(content)
        if creation_date:
            # Set the creation and modification date of the file
            ts = datetime.combine(creation_date, datetime.min.time()).timestamp()
            os.utime(fullname, (ts, ts))

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
        length_bytes: t.Optional[int] = None,  # optional length in bytes
    ) -> t.Optional[NativeDirectoryEntry]:
        """
        Create a new file with a given length
        """
        if not fullname.startswith("/") and not fullname.startswith("\\"):
            fullname = os.path.join(self.pwd, fullname)
        if length_bytes is None:
            length_bytes = number_of_blocks * BLOCK_SIZE
        with open(fullname, "wb") as f:
            f.truncate(length_bytes)
        if creation_date:
            # Set the creation and modification date of the file
            ts = datetime.combine(creation_date, datetime.min.time()).timestamp()
            os.utime(fullname, (ts, ts))
        return NativeDirectoryEntry(fullname)

    def create_directory(
        self,
        fullname: str,
        options: t.Dict[str, t.Union[bool, str]],
    ) -> t.Optional["NativeDirectoryEntry"]:
        """
        Create a directory
        """
        if not fullname.startswith("/") and not fullname.startswith("\\"):
            fullname = os.path.join(self.pwd, fullname)
        fullname = os.path.normpath(fullname)
        os.makedirs(os.path.join(self.base, fullname))  # May raise OSError in case of errors
        return NativeDirectoryEntry(os.path.join(self.base, fullname))

    def chdir(self, fullname: str) -> bool:
        if not fullname.startswith("/") and not fullname.startswith("\\"):
            fullname = os.path.join(self.pwd, fullname)
        fullname = os.path.normpath(fullname)
        if os.path.isdir(os.path.join(self.base, fullname)):
            self.pwd = fullname
            # Change the current working directory
            os.chdir(os.path.join(self.base, fullname))
            return True
        else:
            return False

    def isdir(self, fullname: str) -> bool:
        """
        Check if the path is a directory
        """
        if not fullname.startswith("/") and not fullname.startswith("\\"):
            fullname = os.path.join(self.pwd, fullname)
        return os.path.isdir(os.path.join(self.base, fullname))

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        if options.get("brief"):
            # Lists only file names and file types
            for x in self.filter_entries_list(pattern):
                sys.stdout.write(f"{x.basename}\n")
        else:
            for x in self.filter_entries_list(pattern):
                et = f"{format_size(x.get_size()):>15s}" if x.is_regular_file else x.entry_type
                dt = x.creation_date.strftime("%d-%b-%Y %H:%M ") if x.creation_date else ""
                sys.stdout.write(f"{et:>15s} {dt:>19s} {x.basename}\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        pass

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        stat = os.statvfs(self.base)
        return stat.f_frsize * stat.f_blocks

    def close(self) -> None:
        pass

    def get_pwd(self) -> str:
        return self.pwd

    def __str__(self) -> str:
        return self.base



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import math
import os
import sys
import typing as t
from datetime import date, datetime

from ..abstract import AbstractDirectoryEntry, AbstractFile, AbstractFilesystem
from ..commons import ASCII, IMAGE, READ_FILE_FULL, dump_struct, filename_match
from ..device.abstract import AbstractDevice
from ..unix.commons import unix_join, unix_split
from .dgdosfs import (
    ATCHA,
    ATCON,
    ATDIR,
    ATLNK,
    ATPAR,
    ATPER,
    ATRAN,
    ATRP,
    ATWP,
    format_attr,
    rdos_canonical_filename,
    rdos_join,
    rdos_to_date,
)

__all__ = [
    "DGDOSDumpFile",
    "DGDOSDumpFilesystem",
]


# RDOS System Reference - DUMP File format Pag 125
# https://bitsavers.trailing-edge.com/pdf/dg/software/rdos/093-400027-00_DGDOSDump_SystemReference_Oct83.pdf

NAME_BLOCK_ID = 0o377  # Name block
DATA_BLOCK_ID = 0o376  # Data block
ERROR_BLOCK_ID = 0o375  # Error block
END_BLOCK_ID = 0o374  # End block
TIME_BLOCK_ID = 0o373  # Time block
LINK_DATA_BLOCK_ID = 0o372  # Link data block
LINK_ACCESS_ATTRIBUTES_BLOCK_ID = 0o371  # Link access attributes block
END_OF_SEGMENT_BLOCK_ID = 0o370  # End of segment block


class AbstractBlock:
    """
    Every dumpfile begins with a Name Block and ends with an End Block.
    Each subdirectory in the dumpfile also ends with an end block.
    """

    fs: "DGDOSDumpFilesystem"
    position: int

    def __init__(self, fs: "DGDOSDumpFilesystem", position: int):
        self.fs = fs
        self.position = position

    @classmethod
    def read(cls, fs: "DGDOSDumpFilesystem") -> "AbstractBlock":
        position = fs.f.tell()
        block_id = fs.read_byte()
        if block_id == NAME_BLOCK_ID:
            return NameBlock(fs, position)
        elif block_id == DATA_BLOCK_ID:
            return DataBlock(fs, position)
        elif block_id == ERROR_BLOCK_ID:
            return ErrorBlock(fs, position)
        elif block_id == END_BLOCK_ID:
            return EndBlock(fs, position)
        elif block_id == TIME_BLOCK_ID:
            return TimeBlock(fs, position)
        elif block_id == LINK_DATA_BLOCK_ID:
            return LinkDataBlock(fs, position)
        elif block_id == LINK_ACCESS_ATTRIBUTES_BLOCK_ID:
            return LinkAccessAttributesBlock(fs, position)
        elif block_id == END_OF_SEGMENT_BLOCK_ID:
            return EndOfSegmentBlock(fs, position)
        else:
            raise Exception(f"{block_id} is not a valid block type")


class NameBlock(AbstractBlock):
    """
    Name block
    """

    attributes: int  # File attributes
    contiguous: int  # Contiguous block number
    data: bytes  # File name

    def __init__(self, fs: "DGDOSDumpFilesystem", position: int):
        super().__init__(fs, position)
        self.attributes = fs.read_word()
        if self.attributes & ATCON != 0:
            self.contiguous = fs.read_word()
        self.data = fs.read_to_null()


class DataBlock(AbstractBlock):
    """
    Data block
    """

    byte_count: int  # Number of bytes in the block
    checksum: int  # Checksum
    data_position: int  # File position of the data
    data: bytes  # Data

    def __init__(self, fs: "DGDOSDumpFilesystem", position: int):
        super().__init__(fs, position)
        self.byte_count = fs.read_word()
        self.checksum = fs.read_word()
        self.data_position = fs.f.tell()
        self.data = fs.f.read(self.byte_count)


class EndBlock(AbstractBlock):
    """
    End block
    """


class ErrorBlock(AbstractBlock):
    """
    Error block
    """


class TimeBlock(AbstractBlock):
    """
    Time block
    """

    last_access_date: int
    last_modification_date: int
    last_modification_time: int

    def __init__(self, fs: "DGDOSDumpFilesystem", position: int):
        super().__init__(fs, position)
        self.last_access_date = fs.read_word()
        self.last_modification_date = fs.read_word()
        self.last_modification_time = fs.read_word()


class LinkDataBlock(AbstractBlock):
    """
    Link data block
    """

    def __init__(self, fs: "DGDOSDumpFilesystem", position: int):
        super().__init__(fs, position)
        self.dirname = fs.read_to_null()
        self.resfilename = fs.read_to_null()


class LinkAccessAttributesBlock(AbstractBlock):
    """
    Link access attributes block
    """

    def __init__(self, fs: "DGDOSDumpFilesystem", position: int):
        super().__init__(fs, position)
        self.attributes = fs.read_word()


class EndOfSegmentBlock(AbstractBlock):
    """
    End of segment block
    """

    def __init__(self, fs: "DGDOSDumpFilesystem", position: int):
        super().__init__(fs, position)
        self.t = fs.read_word()
        self.segment_number = ord(fs.f.read(1))
        self.filename = fs.read_to_null()


class DGDOSDumpFile(AbstractFile):
    entry: "DGDOSDumpEntry"
    closed: bool

    def __init__(self, entry: "DGDOSDumpEntry", file_mode: t.Optional[str] = None):
        self.entry = entry
        self.closed = False
        self.file_mode = file_mode or IMAGE

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.get_length()
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        data = bytearray()
        for i in range(block_number, block_number + number_of_blocks):
            self.entry.fs.f.seek(self.entry.addresses[i])
            data.extend(self.entry.fs.f.read(self.entry.block_size))
        # data = self.entry.fs.f.read(number_of_blocks * BLOCK_SIZE)
        # Convert to ASCII if needed
        if self.file_mode == ASCII:
            return bytes([0x0A if x == 0x0D else x for x in data])
        else:
            return bytes(data)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.entry.get_length()

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.entry.get_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return self.entry.get_block_size()

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class DGDOSDumpEntry(AbstractDirectoryEntry):
    fs: "DGDOSDumpFilesystem"
    parent: t.Optional["DGDOSDumpEntry"]  # Parent directory/partition
    filename: str = ""
    extension: str = ""
    attributes: int = 0
    link_access_attributes: int = 0
    size: int = 0  # Size in bytes
    block_size: int = 0
    addresses: t.List[int] = []  # List of block file position
    last_access_date: int = 0  # days since 1967-12-31
    last_modification_date: int = 0  # days since 1967-12-31
    last_modification_time: int = 0  # hour (high byte) and minute (low byte)
    target: str = ""  # link target

    def __init__(self, fs: "DGDOSDumpFilesystem", name_block: NameBlock, parent: t.Optional["DGDOSDumpEntry"] = None):
        self.fs = fs
        basename = name_block.data.decode("ascii", "ignore").rstrip("\0")
        try:
            self.filename, self.extension = basename.split(".", 1)
        except Exception:
            self.filename = basename
            self.extension = ""
        self.attributes = name_block.attributes
        self.parent = parent
        self.addresses = []

    @classmethod
    def create(
        cls,
        fs: "DGDOSDumpFilesystem",
        parent: t.Optional["DGDOSDumpEntry"],
        filename: str,
        length: int,  # Length in blocks
        creation_date: t.Optional[t.Union[date, datetime]] = None,  # optional creation date
        length_bytes: t.Optional[int] = None,  # optional length int bytes
    ) -> "DGDOSDumpEntry":
        """
        Create a new regular file
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def write_buffer(self, buffer: bytearray, position: int) -> None:
        """
        Write the UFD entry to the buffer
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    @property
    def is_random(self) -> bool:
        """
        Check if the file is random organized
        """
        return self.attributes & ATRAN != 0

    @property
    def is_contiguous(self) -> bool:
        """
        Check if the file is contiguous organized
        """
        return self.attributes & ATCON != 0

    @property
    def is_sequential(self) -> bool:
        """
        Check if the file is sequential organized
        """
        return not self.is_random and not self.is_contiguous and not self.is_link

    @property
    def is_link(self) -> bool:
        """
        Check if the file is a link
        """
        return self.attributes & ATLNK != 0

    @property
    def is_directory(self) -> bool:
        """
        Check if the file is a directory
        """
        return self.attributes & ATDIR != 0

    @property
    def is_partition(self) -> bool:
        """
        Check if the file is a partition
        """
        return self.attributes & ATPAR != 0

    @property
    def is_empty(self) -> bool:
        return self.filename == "" and self.extension == ""

    @property
    def fullname(self) -> str:
        if self.parent:
            return rdos_join(self.parent.fullname, self.basename)
        else:
            return self.basename

    @property
    def basename(self) -> str:
        return f"{self.filename}.{self.extension}"

    @property
    def last_access(self) -> t.Optional[date]:
        """
        Last access date
        """
        return rdos_to_date(self.last_access_date)

    @property
    def creation_date(self) -> t.Optional[date]:
        """
        Last modification date
        """
        return rdos_to_date(self.last_modification_date, self.last_modification_time)

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return int(math.ceil(self.size / self.block_size))

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.size

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return self.block_size

    def delete(self) -> bool:
        """
        Delete the directory entry
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def write(self) -> bool:
        """
        Write the directory entry
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def examine(self) -> str:
        if self.is_directory:  # Directory
            file_type = "Directory"
        elif self.is_partition:  # Partition
            file_type = "Partition"
        elif self.is_link:  # Link entry
            file_type = "Link"
        elif self.is_random:  # Random organized file
            file_type = "Random file"
        elif self.is_contiguous:  # Contiguous organized file
            file_type = "Contiguous file"
        else:  # Sequential organized file
            file_type = "Sequential file"
        if self.is_link:
            data: t.Dict[str, t.Any] = {
                "Filename": self.fullname,
                "File type": file_type,
                "Creation date": str(self.creation_date),
                "Target": self.target,
            }
        else:
            data = {
                "Filename": self.fullname,
                "File type": file_type,
                "Creation date": str(self.creation_date),
                "Last access": str(self.last_access),
                "Address": self.addresses,
                "File size": f"{self.get_size()}",
                "Write protected": self.attributes & ATWP != 0,
                "Read protected": self.attributes & ATRP != 0,
                "Immutable attribs": self.attributes & ATCHA != 0,
                "Permanent": self.attributes & ATPER != 0,
                "Link attributes": format_attr(self.link_access_attributes),
            }
        return dump_struct(data) + "\n"

    def open(self, file_mode: t.Optional[str] = None) -> DGDOSDumpFile:
        """
        Open a file
        """
        return DGDOSDumpFile(self, file_mode)

    def __str__(self) -> str:
        attr = format_attr(self.attributes, long=True)
        if self.is_link:
            return f"{self.filename:>10s}.{self.extension:<2s} {attr:<12}  -> {self.target}"
        else:
            uftlkl = format_attr(self.link_access_attributes)
            if uftlkl:
                attr = f"{attr}/{uftlkl}"
            creation_date = self.creation_date.strftime("%m/%d/%y") if self.creation_date else ""
            return (
                # f"{self.filename:>10s}.{self.extension:<2s} "
                f"{self.fullname:<30s} "
                f"{attr:<12} "
                f"{self.get_size():>10d}  "
                f"{creation_date:<8}  "
                # f"{self.addresses[0]:>10}"
            )

    def __repr__(self) -> str:
        return str(self)


class DGDOSDumpFilesystem(AbstractFilesystem):
    """
    Data General DOS/RDOS DUMP

    RDOS System Reference - Pag 125
    https://bitsavers.trailing-edge.com/pdf/dg/software/rdos/093-400027-00_DGDOSDump_SystemReference_Oct83.pdf
    """

    fs_name = "dump"
    fs_description = "Data General DOS/RDOS DUMP"
    fs_platforms = ["nova"]

    f: "AbstractFile"
    pwd: str = "/"  # Current working directory

    def __init__(self, file: "AbstractFile"):
        self.f = file

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        return self.f.read_block(block_number, number_of_blocks)

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        **kwargs: t.Union[bool, str],
    ) -> "DGDOSDumpFilesystem":
        if not isinstance(file_or_dev, AbstractFile):
            raise OSError(errno.EIO, "Not a file")
        self = cls(file_or_dev)
        if strict:
            self.f.seek(0)
            if not self.read_byte() == NAME_BLOCK_ID:
                raise OSError(errno.EIO, "Invalid dump file")
        return self

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
    ) -> t.Iterator["DGDOSDumpEntry"]:
        if pattern:
            pattern = pattern.upper()
        if not pattern and expand:
            pattern = "*"
        if pattern and pattern.startswith("/"):
            absolute_path = pattern
        else:
            absolute_path = unix_join(self.pwd, pattern or "")
        if self.isdir(absolute_path):
            if not expand:
                yield self.get_file_entry(absolute_path)
                return
            dirname = pattern
            pattern = "*"
        else:
            dirname, pattern = unix_split(absolute_path)
        if dirname == "/":  # Root directory
            dir_ufd: t.Optional[DGDOSDumpEntry] = None
        else:
            dir_ufd = self.get_file_entry(dirname)  # type: ignore
        for entry in self.read_dir_entries(dir_ufd):
            if filename_match(entry.basename, pattern, wildcard):
                yield entry

    def get_ufd(self, parent: t.Optional[DGDOSDumpEntry], basename: str) -> DGDOSDumpEntry:
        """
        Get User File Descriptor for a file in a directory
        """
        basename = rdos_canonical_filename(basename).rstrip(".")
        for entry in self.read_dir_entries(parent):
            if filename_match(entry.basename.rstrip("."), basename, wildcard=False):
                return entry
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), basename)

    @property
    def entries_list(self) -> t.Iterator["DGDOSDumpEntry"]:
        yield from self.read_dir_entries()

    def get_file_entry(self, fullname: str) -> DGDOSDumpEntry:
        """
        Get the directory entry for a file
        """
        fullname = unix_join(self.pwd, fullname) if not fullname.startswith("/") else fullname
        parts = [x for x in fullname.split("/") if x]
        if not parts:
            parts = []
        entry: t.Optional[DGDOSDumpEntry] = None
        for i, part in enumerate(parts):
            if entry is not None and not (entry.is_directory or entry.is_partition):  # is a directory?
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)
            entry = self.get_ufd(entry, part)
        if entry is None:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)
        return entry

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        """
        Write content to a file
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
        length_bytes: t.Optional[int] = None,  # optional length in bytes
    ) -> t.Optional[DGDOSDumpEntry]:
        """
        Create a new file with a given length in number of blocks
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def isdir(self, fullname: str) -> bool:
        return False

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        for x in self.filter_entries_list(pattern, include_all=True, wildcard=True):
            if options.get("brief"):
                # For brief mode, print only the file name
                sys.stdout.write(f"{x.basename}\n")
            elif x.is_link:
                # Print link information
                # filename, target
                sys.stdout.write(f"{x.basename:<13s}             {x.target}\n")
            else:
                # Print file information
                # filename, byte length, attributes, last modification date, last access date
                attr = format_attr(x.attributes)
                uftlkl = format_attr(x.link_access_attributes)
                if uftlkl:
                    attr = f"{attr}/{uftlkl}"
                creation_date = x.creation_date.strftime("%m/%d/%y %H:%M") if x.creation_date else ""
                access_date = x.last_access.strftime("%m/%d/%y") if x.last_access else ""
                sys.stdout.write(
                    f"{x.basename:<13s}{x.get_size():>10d}  {attr:<7} {creation_date:<14} {access_date:<8}\n"
                )
        sys.stdout.write("\n")

    def read_byte(self) -> int:
        """
        Read a byte from the dump file
        """
        buffer = self.f.read(1)
        if len(buffer) == 0:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        return ord(buffer)

    def read_word(self) -> int:
        """
        Read a word from the dump file
        """
        buffer = self.f.read(2)
        if len(buffer) < 2:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        return buffer[0] << 8 | buffer[1]

    def read_to_null(self) -> bytes:
        """
        Read from the dump file until a null byte is found
        """
        buffer = bytearray()
        while True:
            byte = self.f.read(1)
            if not byte or ord(byte) == 0:
                break
            buffer.append(ord(byte))
        return bytes(buffer)

    def read_blocks(self) -> t.Iterator[AbstractBlock]:
        self.f.seek(0)
        while True:
            yield AbstractBlock.read(self)

    def read_dir_entries(self, parent: t.Optional[DGDOSDumpEntry] = None) -> t.Iterator[DGDOSDumpEntry]:
        parents = []
        entry: DGDOSDumpEntry = None  # type: ignore
        for block in self.read_blocks():
            if isinstance(block, NameBlock):
                # Name block - create a new entry
                if entry is not None:
                    # Yield the previous entry
                    position = self.f.tell()  # Save current dump file position
                    if entry.is_directory:
                        parents.append(entry)
                    elif entry.is_partition:
                        parents = [entry]
                    if (parent is None and entry.parent is None) or (
                        (entry.parent is not None)
                        and (parent is not None)
                        and (entry.parent.fullname == parent.fullname)
                    ):
                        yield entry
                    self.f.seek(position)  # Restore dump file position
                entry = DGDOSDumpEntry(self, block, parent=parents[-1] if parents else None)
            elif isinstance(block, DataBlock):
                # Data block - save the block address in the entry
                entry.size += block.byte_count
                if entry.block_size < block.byte_count:
                    entry.block_size = block.byte_count
                entry.addresses.append(block.data_position)
            elif isinstance(block, TimeBlock):
                # Time block - set the last access and modification date in the entry
                entry.last_access_date = block.last_access_date
                entry.last_modification_date = block.last_modification_date
                entry.last_modification_time = block.last_modification_time
            elif isinstance(block, LinkDataBlock):
                # Link data block - set the link target in the entry
                dirname = block.dirname.decode("ascii", "ignore").rstrip("\0")
                resfilename = block.resfilename.decode("ascii", "ignore").rstrip("\0")
                if dirname:
                    entry.target = f"{dirname}:{resfilename}"
                else:
                    entry.target = f"{resfilename}"
            elif isinstance(block, LinkAccessAttributesBlock):
                # Link access attributes block - set the link access attributes in the entry
                entry.link_access_attributes = block.attributes
            elif isinstance(block, EndBlock):
                # End block - end of the directory entry/dump file
                if parents:
                    parents.pop()
                else:
                    break
            # elif isinstance(block, EndOfSegmentBlock):
            #     raise TODO
        # Yield the previous entry
        if entry is not None:
            yield entry

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if not arg:
            for entry in self.read_dir_entries():
                if options.get("full") or not entry.is_empty:
                    sys.stdout.write(f"{entry}\n")
        else:
            # Display the file information
            entry = self.get_file_entry(arg)  # type: ignore
            sys.stdout.write(entry.examine())

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.f.get_size()

    @classmethod
    def initialize(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "AbstractFilesystem":
        """Initialize the filesystem"""
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def close(self) -> None:
        self.f.close()

    def chdir(self, fullname: str) -> bool:
        """
        Change the current directory
        """
        return False

    def get_pwd(self) -> str:
        """
        Get the current directory
        """
        return ""

    def get_types(self) -> t.List[str]:
        """
        Get the list of the supported file types
        """
        return []

    def __str__(self) -> str:
        return str(self.f)



# Copyright (C) 2414 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import io
import math
import os
import sys
import typing as t
from functools import reduce

from ..abstract import AbstractFile
from ..commons import ASCII, IMAGE, READ_FILE_FULL
from ..device.abstract import AbstractDevice
from ..device.block_18bit import (
    BYTES_PER_WORD_18BIT,
    BlockDevice18Bit,
    from_18bit_words_to_bytes,
)
from ..unix.commons import (
    UNIXDirectory,
    UNIXDirectoryEntry,
    UNIXFile,
    UNIXFilesystem,
    UNIXInode,
    unix_join,
)

__all__ = [
    "UNIXFile0",
    "UNIXDirectoryEntry0",
    "UNIX0Filesystem",
]

V0_SUPER_BLOCK = 0  # Superblock

V0_IO_BYTES_PER_WORD = 3  # When files are exported, each word is encoded in 3 bytes
V0_WORDS_PER_BLOCK = 64  # Number of words per block
V0_BLOCK_SIZE = BYTES_PER_WORD_18BIT * V0_WORDS_PER_BLOCK  # Block size (in bytes)

V0_BLOCKS_PER_SURFACE = 8000  # Number of blocks on a surface
V0_FILESYSTEM_SIZE = 6400  # Number of blocks in the filesystem
V0_INODE_BLOCKS = 710  # Number of i-node blocks
V0_FIRST_INODE_BLOCK = 2  # First i-node block number
V0_INODE_SIZE = 12  # Inode size (in words)
V0_INODES_PER_BLOCK = V0_WORDS_PER_BLOCK // V0_INODE_SIZE  # Number of inodes per block
V0_DIRENT_SIZE = 8  # Size of a directory entry (in words)
V0_SURFACE_SIZE = V0_BLOCKS_PER_SURFACE * V0_WORDS_PER_BLOCK * BYTES_PER_WORD_18BIT
V0_FREE_BLOCKS_LIST_SIZE = 9  # Number of free blocks in a block (in words)

V0_MAXINT = 0o777777  # Biggest unsigned integer

V0_FLAGS = 0
V0_ADDR = 1
V0_UID = 8
V0_NLINKS = 9
V0_SIZE = 10
V0_UNIQ = 11

V0_NUMBLKS = 7  # Seven block pointers in i-node

V0_USED = 0o400000  # i-node is allocated
V0_LARGE = 0o200000  # large file
V0_SPECIAL = 0o000040  # special file
V0_DIR = 0o000020  # directory

V0_ROWN = 0o000010  # read, owner
V0_WOWN = 0o000004  # write, owner
V0_ROTH = 0o000002  # read, non-owner
V0_WOTH = 0o000001  # write, non-owner

V0_ROOT_INODE = 4  # 'dd' folder


def get_v0_inode_block_offset(inode_num: int) -> t.Tuple[int, int]:
    """
    Return block number and offset for an inode number
    """
    block_num = V0_FIRST_INODE_BLOCK + (inode_num // V0_INODES_PER_BLOCK)
    offset = V0_INODE_SIZE * (inode_num % V0_INODES_PER_BLOCK)
    return block_num, offset


class UNIXFile0(UNIXFile):
    inode: "UNIXInode0"

    def __init__(self, inode: "UNIXInode0", file_mode: t.Optional[str] = None):
        super().__init__(inode)
        self.file_mode = file_mode or IMAGE

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.inode.get_length()
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.inode.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        data = bytearray()
        # Get the blocks to be read
        blocks = list(self.inode.blocks())[block_number : block_number + number_of_blocks]
        # Read the blocks
        for disk_block_number in blocks:
            words = self.inode.fs.read_words_block(V0_BLOCKS_PER_SURFACE + disk_block_number)
            t = from_18bit_words_to_bytes(words, self.file_mode)
            data.extend(t)
        return bytes(data)

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        if self.file_mode == ASCII:
            return self.inode.size * 2  # 2 ASCII bytes per 18 bit word
        else:
            return self.inode.size * V0_IO_BYTES_PER_WORD


class UNIXInode0(UNIXInode):
    """
    Inode numbers begin at 1, and the storage for i-nodes begins at block 2.
    Blocks 2 to 711 contain the i-nodes, with five 12-word i-nodes per block.

    Reserved i-node numbers:

     1  The core file
     2  The "dd" directory
     3  The "system" directory
     6  The "ttyin" special file
     7  The "keyboard" GRAPHIC-2 keyboard special file
     8  The "pptin" paper tape reader special file
    10  The "ttyout" special file
    11  The "display" GRAPHIC-2 display special file
    12  The "pptout" paper tape punch special file

    https://github.com/DoctorWkt/pdp7-unix/blob/master/man/fs.5
    """

    fs: "UNIX0Filesystem"
    uniq: int  # Unique value assigned at creation
    inode_num: int  # Inode number
    flags: int  # Flags
    uid: int  # Owner user id
    nlinks: int  # Link count
    size: int  # Size (in words)
    addr: t.List[int]  # Indirect blocks or data blocks

    @classmethod
    def read(cls, fs: "UNIX0Filesystem", inode_num: int, words: t.List[int], position: int = 0) -> "UNIXInode0":  # type: ignore
        self = UNIXInode0(fs)
        self.inode_num = inode_num
        self.flags = words[position + V0_FLAGS]
        self.uid = words[position + V0_UID]  # Owner user id
        if self.uid == V0_MAXINT:
            self.uid = -1  # 'system' (root) uid
        self.nlinks = V0_MAXINT - words[position + V0_NLINKS] + 1  # Link count
        self.size = words[position + V0_SIZE]  # Size (in words)
        self.uniq = words[position + V0_UNIQ]  # Unique value assigned at creation
        self.addr = words[position + V0_ADDR : position + V0_ADDR + V0_NUMBLKS]  # Indirect blocks or data blocks
        return self

    def write(self) -> None:
        """
        Write inode
        """
        block_number, offset = get_v0_inode_block_offset(self.inode_num)
        words = self.fs.read_words_block(V0_BLOCKS_PER_SURFACE + block_number)
        words[offset : offset + V0_INODE_SIZE] = self.to_words()
        self.fs.write_words_block(V0_BLOCKS_PER_SURFACE + block_number, words)

    def to_words(self) -> t.List[int]:
        """
        Convert inode data to 18bit words
        """
        words = [0] * V0_INODE_SIZE
        words[V0_FLAGS] = self.flags
        if self.uid == -1:
            words[V0_UID] = V0_MAXINT
        else:
            words[V0_UID] = self.uid
        words[V0_NLINKS] = V0_MAXINT - self.nlinks + 1
        words[V0_SIZE] = self.size
        words[V0_UNIQ] = self.uniq
        words[V0_ADDR : V0_ADDR + V0_NUMBLKS] = self.addr
        return words

    def blocks(self, include_indexes: bool = False) -> t.Iterator[int]:
        if self.is_large:
            # Large file
            for block_number in self.addr:
                if block_number == 0:
                    break
                if include_indexes:
                    yield block_number
                for n in self.fs.read_words_block(V0_BLOCKS_PER_SURFACE + block_number):
                    if n == 0:
                        break
                    yield n
        else:
            # Small file
            for block_number in self.addr:
                if block_number == 0:
                    break
                yield block_number

    @property
    def isdir(self) -> bool:
        return (self.flags & V0_DIR) == V0_DIR

    @property
    def is_special_file(self) -> bool:
        return (self.flags & V0_SPECIAL) == V0_SPECIAL

    @property
    def is_regular_file(self) -> bool:
        return not self.isdir

    @property
    def is_large(self) -> bool:
        return bool(self.flags & V0_LARGE)

    @property
    def is_allocated(self) -> bool:
        return (self.flags & V0_USED) != 0

    def read_words(self) -> t.List[int]:
        """
        Read inode data as 18bit words
        """
        data = []
        for block_number in self.blocks():
            data.extend(self.fs.read_words_block(V0_BLOCKS_PER_SURFACE + block_number))
        return data

    def get_block_size(self) -> int:
        """
        Get block size in bytes
        """
        return V0_BLOCK_SIZE

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.size * V0_IO_BYTES_PER_WORD

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return len(list(self.blocks()))

    def examine(self) -> str:
        buf = io.StringIO()
        buf.write("\n*Inode\n")
        buf.write(f"Inode number:          {self.inode_num:>6}\n")
        buf.write(f"Uniq:                  {self.uniq:>6}\n")
        buf.write(f"Flags:                 {self.flags:>06o}\n")
        if self.isdir:
            buf.write("Type:               directory\n")
        elif self.is_special_file:
            buf.write("Type:            special file\n")
        elif self.is_large:
            buf.write("Type:              large file\n")
        else:
            buf.write("Type:                    file\n")
        buf.write(f"Owner user id:         {self.uid:>6}\n")
        buf.write(f"Link count:            {self.nlinks:>6}\n")
        buf.write(f"Size (words):          {self.size:>6}\n")
        if self.is_large:
            buf.write(f"Indirect blocks:       {self.addr}\n")
        buf.write(f"Blocks:                {list(self.blocks())}\n")
        return buf.getvalue()

    def __str__(self) -> str:
        if not self.is_allocated:
            return f"{self.inode_num:>4}# ---"
        else:
            return f"{self.inode_num:>4}# uid: {self.uid:>3}  nlinks: {self.nlinks:>3}  size: {self.size:>5} words  flags: {self.flags:o}"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, UNIXInode0):
            return False
        return (
            self.inode_num == other.inode_num
            and self.uniq == other.uniq
            and self.flags == other.flags
            and self.uid == other.uid
            and self.nlinks == other.nlinks
            and self.size == other.size
            and self.addr == other.addr
        )


class UNIXDirectoryEntry0(UNIXDirectoryEntry):
    inode: "UNIXInode0"

    def open(self, file_mode: t.Optional[str] = None) -> UNIXFile:
        """
        Open a file
        """
        return UNIXFile0(self.inode, file_mode)


class UNIX0Directory(UNIXDirectory):

    @classmethod
    def read(cls, fs: "UNIXFilesystem", inode: "UNIXInode") -> "UNIXDirectory":
        """
        Read the directory
        """
        self = UNIX0Directory(fs, inode)
        if self.inode.isdir:
            self.entries = []
            data = inode.read_words()  # type: ignore
            for i in range(0, len(data), V0_DIRENT_SIZE):
                inum = data[i]
                name = from_18bit_words_to_bytes(data[i + 1 : i + 2 + 4])
                if inum > 0:
                    name_ascii = name.decode("ascii", errors="ignore").rstrip(" \x00")
                    self.entries.append((inum, name_ascii))
        return self

    def write(self) -> None:
        """
        Write the directory
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))


class UNIX0FreeStorageMap:
    """
    Free-Storage Map

    Each block in the free-storage map is structured as follows:

    - the first word is the block number of the next block
      in the free-storage map, or zero if this is the end of
      the free-storage map;
    - the next 9 words hold free block numbers,
      or 0 (no block number).
    """

    fs: "UNIX0Filesystem"
    map_blocks: t.List[int]  # List of map blocks
    free_blocks: t.List[int]  # List of free blocks

    def __init__(self, fs: "UNIX0Filesystem"):
        self.fs = fs
        self.map_blocks = []
        self.free_blocks = []

    @classmethod
    def read(cls, fs: "UNIX0Filesystem") -> "UNIX0FreeStorageMap":
        self = UNIX0FreeStorageMap(fs)
        block_number = fs.storage_map_block
        while block_number != 0:
            self.map_blocks.append(block_number)
            words = fs.read_words_block(V0_BLOCKS_PER_SURFACE + block_number)
            block_number = words[0]
            self.free_blocks.extend(words[1 : V0_FREE_BLOCKS_LIST_SIZE + 1])
        return self

    def write(self) -> None:
        # Check if the map blocks are enough
        required_map_blocks = math.ceil(len(self.free_blocks) / V0_FREE_BLOCKS_LIST_SIZE)
        for _ in range(len(self.map_blocks), required_map_blocks):
            # Allocate a new block for the map
            self.map_blocks.append(self.allocate(1)[0])
        tmp = list(self.free_blocks)
        for i, map_block_number in enumerate(self.map_blocks):
            # Write the map block
            words = [0] * V0_WORDS_PER_BLOCK
            words[0] = self.map_blocks[i + 1] if i + 1 < len(self.map_blocks) else 0
            for j in range(1, V0_FREE_BLOCKS_LIST_SIZE + 1):
                if tmp:
                    words[j] = tmp.pop(0)
                else:
                    words[j] = 0
            self.fs.write_words_block(V0_BLOCKS_PER_SURFACE + map_block_number, words)

    def is_free(self, block_number: int) -> bool:
        """
        Check if the block is free
        """
        return block_number in self.free_blocks

    def set_free(self, block_number: int) -> None:
        """
        Mark the block as free
        """
        # Check if the block is already free
        if block_number not in self.free_blocks:
            # Update the free block list
            for i in range(0, len(self.free_blocks)):
                if self.free_blocks[i] == 0:
                    self.free_blocks[i] = block_number
                    return
            # If there is no free block, add it to the end of the list
            self.free_blocks.append(block_number)

    def set_used(self, block_number: int) -> None:
        """
        Mark the block as used
        """
        # Check if the block is already used
        if block_number in self.free_blocks:
            # Update the free block list
            for i in range(0, len(self.free_blocks)):
                if self.free_blocks[i] == block_number:
                    self.free_blocks[i] = 0
                    break

    def allocate(self, size: int) -> t.List[int]:
        """
        Allocate blocks
        """
        blocks = []
        for block_number in self.free_blocks:
            if block_number != 0:
                blocks.append(block_number)
                self.set_used(block_number)
            if len(blocks) == size:
                break
        if len(blocks) < size:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        return blocks

    def used(self) -> int:
        """
        Count the number of used blocks
        """
        return V0_FILESYSTEM_SIZE - self.free()

    def free(self) -> int:
        """
        Count the number of free blocks
        """
        return len([x for x in self.free_blocks if x != 0])

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, UNIX0FreeStorageMap):
            return False
        return self.map_blocks == other.map_blocks and self.free_blocks == other.free_blocks


class UNIX0Filesystem(UNIXFilesystem):
    """
    UNIX version 0 Filesystem
    """

    fs_name = "unix0"
    fs_description = "UNIX version 0"
    fs_platforms = ["pdp-7"]
    version: int = 0
    directory_class = UNIX0Directory
    storage_map_block: int = 0  # First block of the free-storage map
    dev: BlockDevice18Bit

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"]):
        if isinstance(file_or_device, AbstractFile):
            self.dev = BlockDevice18Bit(file_or_device, words_per_block=V0_WORDS_PER_BLOCK)
        elif isinstance(file_or_device, BlockDevice18Bit):
            self.dev = file_or_device
        else:
            raise OSError(errno.EIO, f"Invalid device type for {self.fs_description} filesystem")

    @classmethod
    def mount(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "UNIX0Filesystem":
        self = cls(file_or_dev)
        self.version = 0
        self.pwd = "/"
        self.inode_size = V0_INODE_SIZE
        self.root_inode = V0_ROOT_INODE
        self.read_superblock()
        return self

    def read_words_block(
        self,
        block_number: int,
    ) -> t.List[int]:
        """
        Read a 256 bytes block as 18bit words
        """
        return self.dev.read_words_block(block_number)

    def write_words_block(
        self,
        block_number: int,
        words: t.List[int],
    ) -> None:
        """
        Write 256 18bit words as a block
        """
        self.dev.write_words_block(block_number, words)

    def read_superblock(self) -> None:
        """Read superblock"""
        # The first word of block 0 points to the first block of the free-storage map.
        words = self.read_words_block(V0_BLOCKS_PER_SURFACE + V0_SUPER_BLOCK)
        self.storage_map_block = words[0]

    def read_inode(self, inode_num: int) -> UNIXInode:
        """
        Read inode by number
        """
        block_number, offset = get_v0_inode_block_offset(inode_num)
        words = self.read_words_block(V0_BLOCKS_PER_SURFACE + block_number)[offset : offset + V0_INODE_SIZE]
        return UNIXInode0.read(self, inode_num, words)

    def read_dir_entries(self, dirname: str) -> t.Iterator["UNIXDirectoryEntry"]:
        inode: UNIXInode0 = self.get_inode(dirname)  # type: ignore
        if inode:
            for inode_num, filename in self.list_dir(inode):
                fullname = unix_join(dirname, filename)
                yield UNIXDirectoryEntry0(self, fullname, inode_num)

    def get_file_entry(self, fullname: str) -> UNIXDirectoryEntry:
        inode: UNIXInode0 = self.get_inode(fullname)  # type: ignore
        if not inode:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)
        return UNIXDirectoryEntry0(self, fullname, inode.inode_num, inode)

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        entries = sorted(self.filter_entries_list(pattern, include_all=True, wildcard=True))
        if not entries:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), pattern)
        if not options.get("brief") and not self.version == 0:
            blocks = reduce(lambda x, y: x + y, [x.inode.get_length() for x in entries])
            if self.version < 3:
                sys.stdout.write(f"total {blocks:>4}\n")
            else:
                sys.stdout.write(f"blocks = {blocks}\n")
        for x in entries:
            if not options.get("full") and x.basename.startswith("."):
                pass
            elif options.get("brief"):
                # Lists only file names
                sys.stdout.write(f"{x.basename}\n")
            uid = x.inode.uid if x.inode.uid != -1 else 0o77
            sys.stdout.write(
                f"{x.inode_num:>03o} {x.inode.flags & 0o77:02o} {uid:02o} {x.inode.nlinks:>02o} {x.inode.size:>05o} {x.basename}\n"
            )



import pytest

from xferx.pdp11.dos11fs import DOS11Filesystem, UserFileDirectoryBlock
from xferx.shell import Shell

DSK = "tests/dsk/dos11_dectape.tap"


def test_dos11_dectape():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /dos11 {DSK}", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, DOS11Filesystem)

    shell.onecmd("dir t:[*,*]", batch=True)
    shell.onecmd("dir /uic t:", batch=True)
    shell.onecmd("type t:1.txt", batch=True)

    x = fs.read_bytes("1000.txt")
    x = x.rstrip(b"\0")
    assert len(x) == 44000
    for i in range(0, 1000):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x

    l = list(fs.entries_list)
    assert len(l) == 9


def test_dos11_dectape_bitmap():
    shell = Shell(verbose=True)
    shell.onecmd(f"copy {DSK} {DSK}.mo", batch=True)
    shell.onecmd(f"mount t: /dos11 {DSK}.mo", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, DOS11Filesystem)

    d = fs.get_file_entry("[1,1]500.TXT")
    assert d is not None
    assert not d.contiguous

    # Write UFD
    d.ufd_block.write()
    ufd_block2 = UserFileDirectoryBlock.read(d.ufd_block.fs, d.ufd_block.uic, d.ufd_block.block_number)
    assert str(d.ufd_block) == str(ufd_block2)

    # Delete linked file
    d.delete()
    with pytest.raises(FileNotFoundError):
        fs.get_file_entry("[200,200]500.TXT")

    # UIC not found
    with pytest.raises(Exception):
        shell.onecmd("copy /TYPE:CONTIGUOUS t:10.TXT t:[123,321]10NEW.TXT", batch=True)

    # Create a contiguous file
    shell.onecmd("copy /TYPE:CONTIGUOUS t:10.TXT t:10NEW.TXT", batch=True)
    x2 = fs.read_bytes("10NEW.txt")
    x2 = x2.rstrip(b"\0")
    assert len(x2) == 440
    for i in range(0, 10):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x2

    # Create a non-contiguous file
    shell.onecmd("copy /TYPE:NOCONTIGUOUS t:10.TXT t:10NEW2.TXT", batch=True)
    x2 = fs.read_bytes("10NEW2.txt")
    x2 = x2.rstrip(b"\0")
    assert len(x2) == 440
    for i in range(0, 10):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x2


def test_dos11_init():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount in: /dos11 {DSK}", batch=True)
    shell.onecmd(f"create /allocate:576 {DSK}.mo", batch=True)
    shell.onecmd(f"init /dos11 /dev:dectape {DSK}.mo", batch=True)
    shell.onecmd(f"mount ou: /dos11 {DSK}.mo", batch=True)
    shell.onecmd("dir ou:", batch=True)
    shell.onecmd("copy in:* ou:", batch=True)
    fs = shell.volumes.get('OU')
    assert isinstance(fs, DOS11Filesystem)

    # Create the UIC
    shell.onecmd("create /directory ou:[10,20]", batch=True)
    with pytest.raises(Exception):
        shell.onecmd("create /directory ou:[10,20]", batch=True)
    shell.onecmd("create ou:[10,20]test /allocate:5", batch=True)
    fs.get_file_entry("[10,20]test")
    shell.onecmd("dir ou:[10,20]", batch=True)



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import os
import sys
import typing as t
from datetime import date

from ..abstract import AbstractDirectoryEntry, AbstractFile, AbstractFilesystem
from ..commons import ASCII, IMAGE, READ_FILE_FULL, filename_match
from ..device.abstract import AbstractDevice
from ..device.block_18bit import BlockDevice18Bit, from_18bit_words_to_bytes
from .adssfs import (
    DECTAPE_BLOCKS,
    WORDS_PER_BLOCK,
    ADSSFilesystem,
    adss_canonical_filename,
    ascii_to_sixbit,
    decode_block_format,
    oct_dump,
    sixbit_to_ascii,
)

__all__ = [
    "DOS15DirectoryEntry",
    "DOS15File",
    "DOS15Filesystem",
]

# Pag 87
# https://bitsavers.org/pdf/dec/pdp15/DEC-15-ODFFA-B-D_DOS-15_System_Manual_197408.pdf

RF_MFD_BLOCK = 0o1777  # MFD block number on RF disk
RP_MFD_BLOCK = 0o47040  # MFD block number on RP disk


def dos15_to_date(val: int) -> t.Optional[date]:
    """
    Translate DOS-15 date to Python date

    Month (bits 0-5), Day (bits 6-11), Year (bits 12-17 module 1970)
    """
    if val == 0:
        return None
    month = val >> 12
    day = (val >> 6) & 0o37
    year = (val & 0o37) + 1970
    try:
        return date(year, month, day)
    except ValueError:
        return None


def dos15_split_fullname(uic: str, fullname: t.Optional[str], wildcard: bool = True) -> t.Tuple[str, t.Optional[str]]:
    """
    Split a fullname into UIC and filename

    UIC and filename are separated by a ":"
    Example: "TMP:FILENAM;EXT"
    """
    if fullname:
        if ":" in fullname:
            try:
                uic, fullname = fullname.split(":", 1)
            except Exception:
                return uic.upper(), fullname
        if fullname:
            fullname = adss_canonical_filename(fullname, wildcard=wildcard)
    return uic.upper(), fullname


class DOS15File(AbstractFile):
    entry: "DOS15DirectoryEntry"
    file_mode: str  # ASCII of IMAGE
    closed: bool

    def __init__(self, entry: "DOS15DirectoryEntry", file_mode: t.Optional[str] = None):
        self.entry = entry
        self.file_mode = file_mode or IMAGE
        self.closed = False

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.get_length()
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        data = bytearray()
        for i, next_block_number in enumerate(self.entry.get_blocks()):
            if i >= block_number:
                words = self.entry.fs.read_words_block(next_block_number)
                t = from_18bit_words_to_bytes(words, self.file_mode)
                data.extend(t)
                number_of_blocks -= 1
                if number_of_blocks == 0:
                    break
        return bytes(data)

    def read_words_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> t.List[int]:
        """
        Read block(s) of words from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.get_length()
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        words: t.List[int] = []
        for i, next_block_number in enumerate(self.entry.get_blocks()):
            if i >= block_number:
                data = self.entry.fs.read_words_block(next_block_number)
                words.extend(data)
                number_of_blocks -= 1
                if number_of_blocks == 0:
                    break
        return words

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def write_words_block(
        self,
        words: t.List[int],
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.entry.get_length()

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.get_length() * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return (WORDS_PER_BLOCK - 2) * 3  # Exclude the last two words (previous/next block number)

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class DOS15DirectoryEntry(AbstractDirectoryEntry):
    """
    Directory Entry

    Word

        +-----------------------------------+
      0 | File name                         |
      1 |                                   |
        +-----------------------------------+
      2 | Extension                         |
        +-----------------------------------+
      3 | Truncated bit | First block       |
        +-----------------------------------+
      4 | Number of blocks                  |
        +-----------------------------------+
      5 | First Retrieval Information Block |
        +-----------------------------------+
      6 | Prot. | RIB first word            |
        +-----------------------------------+
      7 | Creation date                     |
        +-----------------------------------+

    Pag 87
    https://bitsavers.org/pdf/dec/pdp15/DEC-15-ODFFA-B-D_DOS-15_System_Manual_197408.pdf
    """

    directory: "UserFileDirectoryBlock"
    file_number: int = 0  # File number in the directory
    filename: str  # Filename
    extension: str = ""  # File extension
    block_number: int = 0  # First file block number
    raw_data_link: int = 0  # Raw data link (truncated bit, block number)
    length: int  # Number of blocks in the file
    protection_code: int = 0  # Protection code
    retrieval_information_block: int  # RIB first block number
    rib_position: int = 0  # Position in RIB block
    raw_creation_date: int

    def __init__(self, directory: "UserFileDirectoryBlock"):
        self.directory = directory

    @classmethod
    def read(
        cls, directory: "UserFileDirectoryBlock", words: t.List[int], file_number: int, position: int
    ) -> "DOS15DirectoryEntry":
        self = cls(directory)
        self.file_number = file_number
        self.filename = sixbit_to_ascii(words[position]) + sixbit_to_ascii(words[position + 1])
        self.extension = sixbit_to_ascii(words[position + 2])
        self.raw_data_link = words[position + 3]
        self.block_number = self.raw_data_link & 0o377777  # First block number
        self.length = words[position + 4]  # Number of blocks in the file
        self.retrieval_information_block = words[position + 5]
        self.protection_code = words[position + 6] >> 16  # Protection code
        self.rib_position = words[position + 6] & ~0o600000  # RIB first word position
        self.raw_creation_date = words[position + 7]
        return self

    @property
    def is_active(self) -> bool:
        """
        Is the file active?
        """
        return self.block_number != 0 and self.raw_data_link & (1 << 17) == 0

    @property
    def fullname(self) -> str:
        return self.basename

    @property
    def basename(self) -> str:
        return f"{self.filename};{self.extension}"

    def get_blocks(self) -> t.List[int]:
        """
        Get the blocks used by the file
        """
        next_block_number = self.retrieval_information_block
        length = self.length  # Length in blocks
        position = self.rib_position
        blocks = []
        while next_block_number != 0o777777 and length > 0:
            print(next_block_number)
            rib = RetrievalInformationBlock.read(self, next_block_number)
            blocks = rib.data_blocks[position : position + length]
            length -= len(blocks)
            next_block_number = rib.next_block_number
            position = 0
        return blocks

    # def get_file_bitmap(self) -> "ADSSBitmap":
    #     """
    #     The File Bitmap specifies the blocks occupied by the file
    #     """
    #     return ADSSBitmap.read(self.directory.fs, self.file_number)

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.length

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.get_length() * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return (WORDS_PER_BLOCK - 1) * 3

    @property
    def creation_date(self) -> t.Optional[date]:
        return dos15_to_date(self.raw_creation_date)

    def delete(self) -> bool:
        """
        Delete the directory entry
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def write(self) -> bool:
        """
        Write the directory entry
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def open(self, file_type: t.Optional[str] = None) -> DOS15File:
        """
        Open a file
        """
        return DOS15File(self, file_type)

    def read_bytes(self, file_mode: t.Optional[str] = None) -> bytes:
        """Get the content of the file"""
        if file_mode is None:
            file_mode = IMAGE
        # Always read the file as IMAGE
        f = self.open(IMAGE)
        try:
            length = self.get_length()  # Number of blocks in the file
            result = bytearray()
            if file_mode == ASCII:
                for i in range(length):
                    # Read the block, exclude the last two words (previous/next block number)
                    words = f.read_words_block(i)[:-2]
                    result += decode_block_format(words)
            else:
                for i in range(length):
                    # Read the block, exclude the last two words (previous/next block number)
                    words = f.read_words_block(i)[:-2]
                    result += from_18bit_words_to_bytes(words, IMAGE)
            return bytes(result)
        finally:
            f.close()

    @property
    def fs(self) -> "DOS15Filesystem":
        return self.directory.fs

    def __str__(self) -> str:
        return f"{self.file_number:>2}  {self.filename:<6} {self.extension:<3}  {'Y' if self.is_active else 'N'}  {self.block_number:4}  {self.length:4}"


class RetrievalInformationBlock:
    """
    Retrieval Information Block (RIB)

    Word

        +-----------------------------------+
      0 | Number of blocks                  |
        +-----------------------------------+
      1 | Data block 1                      |
        //                                  /
      n | Data block n                      |
        +-----------------------------------+
    254 | Previous RIB block number         |
        +-----------------------------------+
    255 | Next RIB block number             |
        +-----------------------------------+

    Pag 83
    https://bitsavers.org/pdf/dec/pdp15/DEC-15-ODFFA-A_DOS15_SysMan.pdf
    """

    entry: "DOS15DirectoryEntry"
    # Total number of blocks described by this RIB
    number_of_blocks: int = 0
    # List of entries in the UFD block
    data_blocks: t.List[int]
    # Block number of this Master File Directory block
    block_number: int = 0
    # Block number of the next Master File Directory block
    next_block_number: int = 0
    # Block number of the previous Master File Directory block
    previous_block_number: int = 0

    def __init__(self, entry: "DOS15DirectoryEntry"):
        self.entry = entry

    @classmethod
    def read(cls, entry: "DOS15DirectoryEntry", block_number: int) -> "RetrievalInformationBlock":
        self = cls(entry)
        self.block_number = block_number
        # Read the block
        words = self.fs.read_words_block(block_number)
        self.number_of_blocks = words[0]  # Number of blocks in the RIB
        self.data_blocks = words[1 : self.number_of_blocks + 1]  # Data blocks
        self.previous_block_number = words[-2]  # Pointer to previous MFD block
        self.next_block_number = words[-1]  # Pointer to next MFD block
        return self

    @property
    def fs(self) -> "DOS15Filesystem":
        return self.entry.directory.fs


class UserFileDirectoryBlock(object):
    """
    User File Directory Block
    """

    ufd: "MasterFileDirectoryEntry"
    # List of entries in the UFD block
    entries_list: t.List[DOS15DirectoryEntry]
    # Block number of this Master File Directory block
    block_number: int = 0
    # Block number of the next Master File Directory block
    next_block_number: int = 0
    # Block number of the previous Master File Directory block
    previous_block_number: int = 0

    def __init__(self, ufd: "MasterFileDirectoryEntry"):
        self.ufd = ufd

    @classmethod
    def read(cls, ufd: "MasterFileDirectoryEntry", block_number: int) -> "UserFileDirectoryBlock":
        self = cls(ufd)
        self.block_number = block_number
        # Read the UFD block
        self.entries_list = []
        words = self.fs.read_words_block(block_number)
        for i, position in enumerate(range(0, len(words) - 8, 8)):
            entry = DOS15DirectoryEntry.read(self, words, i, position)
            if entry.is_active:
                self.entries_list.append(entry)
        self.previous_block_number = words[-2]  # Pointer to previous MFD block
        self.next_block_number = words[-1]  # Pointer to next MFD block
        return self

    @property
    def fs(self) -> "DOS15Filesystem":
        return self.ufd.fs


class MasterFileDirectoryEntry:
    """
    Master File Directory Entry in the MFD block

    Word

        +-----------------------------------+
      0 | UIC                               |
        +-----------------------------------+
      1 | UFD First Block                   |
        +-----------------------------------+
      2 | Prot. bit | UFD Entry Size        |
        +-----------------------------------+
      3 | Unused                            |
        +-----------------------------------+
    """

    mfd_block: "MasterFileDirectoryBlock"
    uic: str  # User Identification Code
    ufd_block_number: int  # Block number of the file
    protected: bool  # Protected bit
    ufd_entry_size: int  # UFD entry size in bits

    def __init__(self, mfd_block: "MasterFileDirectoryBlock"):
        self.mfd_block = mfd_block

    @classmethod
    def read(
        cls, mfd_block: "MasterFileDirectoryBlock", words: t.List[int], position: int
    ) -> "MasterFileDirectoryEntry":
        """
        Read a Master File Directory entry from the MFD block
        """
        self = cls(mfd_block)
        self.uic = sixbit_to_ascii(words[position])
        self.ufd_block_number = words[position + 1]
        self.protected = bool(words[position + 2] >> 17)
        self.ufd_entry_size = words[position + 2] & ~0o400000
        return self

    def is_active(self) -> bool:
        """
        Check if the entry is active
        """
        return self.ufd_block_number != 0o777777 and self.ufd_entry_size != 0

    def read_ufd_blocks(self) -> t.Iterator["UserFileDirectoryBlock"]:
        """Read User File Directory blocks"""
        next_block_number = self.ufd_block_number
        while next_block_number != 0o777777:
            ufd_block = UserFileDirectoryBlock.read(self, next_block_number)
            next_block_number = ufd_block.next_block_number
            yield ufd_block

    @property
    def entries_list(self) -> t.Iterator[DOS15DirectoryEntry]:
        """
        Iterate over all entries in the User File Directory
        """
        for ufd_block in self.read_ufd_blocks():
            yield from ufd_block.entries_list

    @property
    def fs(self) -> "DOS15Filesystem":
        return self.mfd_block.fs

    def __str__(self) -> str:
        """
        String representation of the Master File Directory entry
        """
        return f"UIC: {self.uic:>3}  Block: {self.ufd_block_number:6d}  Protected: {'Y' if self.protected else 'N'}  Entry Size: {self.ufd_entry_size}"


class MasterFileDirectoryBlock:
    """
    Master File Directory Block


    Word

        +-----------------------------------+
      0 | -1                                |
        +-----------------------------------+
      1 | Bad Allocation Table Block        |
        +-----------------------------------+
      2 | SYS Block                         |
        +-----------------------------------+
      3 | Entry Size | SAT Table Block      |
        +-----------------------------------+
      4 | MFD Entries (4 words each)        |
        /                                   /
        |                                   |
        +-----------------------------------+
    """

    fs: "DOS15Filesystem"
    # List of entries in the MFD block
    entries_list: t.List[MasterFileDirectoryEntry]
    # Bad allocation table first block number
    bad_allocation_table: int
    # SYSBLK first block number
    sysblk: int
    # Storage Allocation Table first block number
    storage_allocation_table: int
    # Block number of this Master File Directory block
    block_number: int = 0
    # Block number of the next Master File Directory block
    next_block_number: int = 0
    # Block number of the previous Master File Directory block
    previous_block_number: int = 0

    def __init__(self, fs: "DOS15Filesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "DOS15Filesystem", block_number: int) -> "MasterFileDirectoryBlock":
        self = cls(fs)
        self.block_number = block_number
        # Read the MFD block
        words = self.fs.read_words_block(RF_MFD_BLOCK)
        if words[0] != 0o777777:
            words = self.fs.read_words_block(RP_MFD_BLOCK)
            assert words[0] == 0o777777
        self.bad_allocation_table = words[1]
        self.sysblk = words[2]
        # MFD entry size in bits, plus the block number of the first submap
        tmp = words[3]
        # entry_size = tmp >> 15
        self.storage_allocation_table = tmp & 0o377777
        self.entries_list = []
        for i in range(4, len(words) - 4, 4):
            entry = MasterFileDirectoryEntry.read(self, words, i)
            if entry.is_active():
                self.entries_list.append(entry)
        self.previous_block_number = words[-2]  # Pointer to previous MFD block
        self.next_block_number = words[-1]  # Pointer to next MFD block
        return self


class MasterFileDirectory:
    """
    Master File Directory

    Pag 81
    https://bitsavers.org/pdf/dec/pdp15/DEC-15-ODFFA-A_DOS15_SysMan.pdf
    """

    fs: "DOS15Filesystem"
    mfd_blocks: t.List[MasterFileDirectoryBlock]  # List of MFD blocks

    def __init__(self, fs: "DOS15Filesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "DOS15Filesystem") -> "MasterFileDirectory":
        """
        Read the Master File Directory (MFD)
        """
        self = cls(fs)
        self.mfd_blocks = []
        mfd_block_number = RF_MFD_BLOCK
        while mfd_block_number != 0o777777:  # 0o777777 is the end of the MFD
            mfd_block = MasterFileDirectoryBlock.read(self.fs, mfd_block_number)
            mfd_block_number = mfd_block.next_block_number
            self.mfd_blocks.append(mfd_block)
        return self

    @property
    def entries_list(self) -> t.Iterator[MasterFileDirectoryEntry]:
        """
        Iterate over all entries in the Master File Directory
        """
        for mfd_block in self.mfd_blocks:
            yield from mfd_block.entries_list

    def get_entry(self, uic: str) -> t.Optional[MasterFileDirectoryEntry]:
        for entry in self.entries_list:
            if entry.uic == uic:
                return entry
        return None


class DOS15Filesystem(AbstractFilesystem):
    """
    DOS-15 Filesystem

    +------------+          +--------+
    |    MFD     +--------> |  SAT   |
    |   Master   |          | Master |
    |    File    |          | Bitmap |
    |  Directory |          +--------+
    +-----+------+         +-----------+    +-----------+
          |                |   UFD     |    |    RIB    |    +---------+
          +--------------> +   User    +--> | Retrieval +--> | Block 1 |
                           |   File    |    |   Block   |    |         |
                           | Directory |    +-----------+    +-------+-+
                           +-----------+                       /|\   |
                                                                |   \|/
                                                             +--+------+
                                                             | Block 2 |
                                                             +-^-----+-+

    Pag 129
    https://bitsavers.org/pdf/dec/pdp15/PDP-15_System_Software_Handouts_1975.pdf
    """

    fs_name = "dos15"
    fs_description = "PDP-15 DOS-15"
    fs_platforms = ["pdp-9", "pdp-15"]
    uic: str = ""  # current User Identification Code (only for disk, not DECtape)
    dev: BlockDevice18Bit

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"]):
        if isinstance(file_or_device, AbstractFile):
            self.dev = BlockDevice18Bit(file_or_device, words_per_block=WORDS_PER_BLOCK)
        elif isinstance(file_or_device, BlockDevice18Bit):
            self.dev = file_or_device
        else:
            raise OSError(errno.EIO, f"Invalid device type for {self.fs_description} filesystem")

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        **kwargs: t.Union[bool, str],
    ) -> t.Union["DOS15Filesystem", "ADSSFilesystem"]:
        """
        Mount the filesystem from a file or device
        """
        self = cls(file_or_dev)
        blocks = self.get_size() // 3 // 512
        is_dectape = abs(blocks - DECTAPE_BLOCKS) < 4
        if is_dectape:
            return ADSSFilesystem.mount(self.dev, strict=strict, **kwargs)

        self.uic = ""
        if strict:
            mfd = MasterFileDirectory.read(self)
            for ufd in mfd.entries_list:
                if not self.uic:
                    self.uic = ufd.uic
                    break
        return self

    def read_words_block(
        self,
        block_number: int,
    ) -> t.List[int]:
        """
        Read a 256 bytes block as 18bit words
        """
        return self.dev.read_words_block(block_number)

    def write_words_block(
        self,
        block_number: int,
        words: t.List[int],
    ) -> None:
        """
        Write 256 18bit words as a block
        """
        self.dev.write_words_block(block_number, words)

    def read_dir_entries(self, uic: t.Optional[str] = None) -> t.Iterator["DOS15DirectoryEntry"]:
        """
        Read directory entries
        """
        mfd = MasterFileDirectory.read(self)
        ufd = mfd.get_entry(uic or self.uic)
        if ufd:
            yield from ufd.entries_list

    @property
    def entries_list(self) -> t.Iterator["DOS15DirectoryEntry"]:
        yield from self.read_dir_entries()

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
        uic: t.Optional[str] = None,
    ) -> t.Iterator["DOS15DirectoryEntry"]:
        uic, pattern = dos15_split_fullname(fullname=pattern, wildcard=wildcard, uic=uic or self.uic)
        for entry in self.read_dir_entries(uic):
            if entry.is_active and filename_match(entry.basename, pattern, wildcard):
                yield entry

    def get_file_entry(self, fullname: str) -> DOS15DirectoryEntry:
        """
        Get the directory entry for a file
        """
        for entry in self.filter_entries_list(fullname, wildcard=False):
            return entry
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
    ) -> t.Optional[DOS15DirectoryEntry]:
        """
        Create a new file with a given length in number of blocks
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    # def read_bitmap(self) -> ADSSBitmap:
    #     """
    #     Read the Directory Bitmap
    #     """

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def isdir(self, fullname: str) -> bool:
        return False

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        """
            10-AUG-71
        MFD DIRECTORY LISTING
          5753 FREE BLKS
            47 USER FILES
           612 USER BLKS
        SCR    115(0)      6   316
        BLD    215(0)     14    23

             06-JAN-79
         DIRECTORY LISTING  (REN)
           6133 FREE BLKS
              6 USER FILES
             26 USER BLKS
        """
        if options.get("uic"):
            # Listing of all UIC
            dt = date.today().strftime('%y-%b-%d').upper()
            sys.stdout.write("UIC\n\n")
            sys.stdout.write(f"     {dt}\n")
            sys.stdout.write(" MFD DIRECTORY LISTING)\n")
            mfd = MasterFileDirectory.read(self)
            for ufd in mfd.entries_list:
                sys.stdout.write(f"{ufd.uic} {ufd}\n")
            return

        uic, pattern = dos15_split_fullname(fullname=pattern, wildcard=True, uic=self.uic)
        entries = list(self.filter_entries_list(pattern, uic=uic, wildcard=True))
        if not entries:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), pattern)
        if not options.get("brief"):
            dt = date.today().strftime('%y-%b-%d').upper()
            used = sum(entry.get_length() for entry in entries if entry.is_active)
            sys.stdout.write(f"     {dt}\n")
            sys.stdout.write(f" DIRECTORY LISTING  ({uic})\n")
            sys.stdout.write(f" {len(entries):>6o} USER FILES\n")
            sys.stdout.write(f" {used:>6o} USER BLKS\n")
        for x in entries:
            if x.is_active:
                if options.get("brief"):
                    sys.stdout.write(f"{x.filename:<6};{x.extension:<3}\n")
                else:
                    creation_date = x.creation_date and x.creation_date.strftime("%d-%b-%y").upper() or ""
                    sys.stdout.write(f" {x.filename:<6} {x.extension:<3}  {x.length:>6o}  {creation_date}\n")
        # if not options.get("brief"):
        #     bitmap = self.read_bitmap()
        #     sys.stdout.write(f"{bitmap.free():<4o} FREE BLOCKS\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        # if options.get("bitmap"):
        #     # Display the bitmap
        #     file_number = int(arg) if arg and arg.isdigit() else None
        #     bitmap = ADSSBitmap.read(self, file_number)
        #     for i in range(0, bitmap.total_bits):
        #         sys.stdout.write(f"{i:>4d} {'[ ]' if bitmap.is_free(i) else '[X]'}  ")
        #         if i % 16 == 15:
        #             sys.stdout.write("\n")
        #     sys.stdout.write(f"\nUsed blocks: {bitmap.used()}\n")
        # elif arg:
        if arg:
            # Display the file entry
            entries = self.filter_entries_list(arg, wildcard=True)
            for entry in entries:
                sys.stdout.write(f"File number:              {entry.file_number}\n")
                sys.stdout.write(f"Filename:                 {entry.filename}\n")
                sys.stdout.write(f"Extension:                {entry.extension}\n")
                sys.stdout.write(f"Raw data link:            {entry.raw_data_link:06o}\n")
                sys.stdout.write(f"Active:                   {'Y' if entry.is_active else 'N'}\n")
                sys.stdout.write(f"First block number:       {entry.block_number}\n")
                sys.stdout.write(f"Size:                     {entry.get_length()}\n")
                sys.stdout.write(f"RIB:                      {entry.retrieval_information_block}\n")
                sys.stdout.write(f"RIB position:             {entry.rib_position}\n")
                sys.stdout.write(f"Creation date:            {entry.creation_date}\n")
                sys.stdout.write(f"Protection code:          {entry.protection_code}\n")
                blocks = str(list(entry.get_blocks()))
                sys.stdout.write(f"Blocks:                   {blocks}\n")
                sys.stdout.write("\n")
        else:
            # Display the directory entries
            sys.stdout.write("Num   Filename  Data link Block  Size  Date\n")
            sys.stdout.write("---   --------  --------- -----  ----  ----\n")
            full = bool(options.get("full", False))
            for entry in self.read_dir_entries():
                if full or entry.is_active:
                    block_number = f"{entry.block_number:4d}" if entry.is_active else "-"
                    length = f"{entry.get_length():4d}" if entry.is_active else "-"
                    sys.stdout.write(
                        f"{entry.file_number:>2}  {entry.filename:>6};{entry.extension:<3}  "
                        f"{entry.raw_data_link:06o} {block_number:>8}  {length:>4}  "
                        f"{entry.creation_date}\n"
                    )

    def dump(self, fullname: t.Optional[str], start: t.Optional[int] = None, end: t.Optional[int] = None) -> None:
        """Dump the content of a file or a range of blocks"""
        if fullname:
            entry = self.get_file_entry(fullname)
            if start is None:
                start = 0
            blocks = entry.get_blocks()
            if end is None or end > len(blocks) - 1:
                end = entry.get_length() - 1
            for block_number in range(start, end + 1):
                words = self.read_words_block(blocks[block_number])
                sys.stdout.write(f"\nBLOCK NUMBER   {blocks[block_number]:08} ({block_number:08})\n")
                oct_dump(words)
        else:
            if start is None:
                start = 0
                if end is None:  # full disk
                    end = (self.get_size() // WORDS_PER_BLOCK // 4) - 1
            elif end is None:  # one single block
                end = start
            for block_number in range(start, end + 1):
                words = self.read_words_block(block_number)
                sys.stdout.write(f"\nBLOCK NUMBER   {block_number:08}\n")
                oct_dump(words)

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    def chdir(self, fullname: str) -> bool:
        """
        Change the current User Identification Code
        """
        mfd = MasterFileDirectory.read(self)
        fullname = sixbit_to_ascii(ascii_to_sixbit(fullname[0:3]))
        entry = mfd.get_entry(fullname)
        if entry is None:
            return False
        self.uic = entry.uic
        return True

    def get_pwd(self) -> str:
        """
        Get the current User Identification Code
        """
        return self.uic
